import ast
from collections.abc import Iterator
import enum
import inspect
import re
import sys
import traceback
from unittest.mock import call
import uuid
import warnings
import astor
from copy import deepcopy
from typing import Callable, NoReturn

from malevich._autoflow.flow import Flow
from malevich._autoflow.tracer import autoflow, traced
from malevich._autoflow.tree import ExecutionTree
from malevich.models.exceptions import NewVariableInBranchError
from .models.mappers.conditional import AddNegativeCondition, AddPositiveCondition
from malevich.models.nodes.operation import OperationNode

from .models.nodes.base import BaseNode
from .models.nodes.morph import MorphNode

State = tuple[dict, dict] # globals, locals

def copy_state(state: State):
    return state[0], deepcopy(state[1])


def extract_conditioned_nodes(old_tree: ExecutionTree, new_tree: ExecutionTree):
    old_tree_nodes = set(old_tree.nodes())
    new_tree_nodes = set(new_tree.nodes())
    new_nodes = new_tree_nodes.difference(old_tree_nodes)

    for new_node in new_nodes:
        for from_node, _, _ in new_tree.edges_to(new_node):
            if from_node not in old_tree_nodes:
                break
        else:
            yield new_node


def prep_synerr_assign(name: str, stmts: list[ast.stmt]) -> NewVariableInBranchError | None:
    for stmt in stmts:
        for node in ast.walk(stmt):
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    for inode in ast.walk(target):
                        if isinstance(inode, ast.Name) and inode.id == name:
                            return NewVariableInBranchError(
                                var_name=name,
                                function_name=None,
                                # FIXME: wrong lineno
                                details=('<string>', stmt.lineno,
                                         target.col_offset, ast.unparse(node))
                            )
    return None


def retrace(tree: ExecutionTree, state):
    for value in state[1].values():
        if isinstance(value, traced):
            value._autoflow._tree_ref = tree

    return state


def merge_detached_tree_inplace(tree: ExecutionTree, detached_tree: ExecutionTree) -> None:
    edges = [*tree.tree, *detached_tree.tree]
    cedges = dict()
    for f, t, l in edges:
        cedges[(f.owner.uuid, t.owner.uuid)] = (f, t, l)

    tree.tree = list(cedges.values())


def combine_branch_states_inplace(state, if_state, else_state, cond_node):
    if else_state is None:
        else_state = state

    initial_state, if_lstate, else_lstate = (
        state[1], if_state[1], else_state[1],
    )

    for key in initial_state:
        if key in if_lstate and key in else_lstate and (
            initial_state[key] is not if_lstate[key]        # key changed in at least
            or initial_state[key] is not else_lstate[key]   # one of the branches
        ):
            v, ifv, elsev = initial_state[key], if_lstate[key], else_lstate[key]
            if (
                all(map(lambda x: isinstance(x, traced) or x is None, (v, ifv, elsev)))
                and all(map(lambda x: isinstance(x.owner, BaseNode) or x is None, (v, ifv, elsev)))
            ):
                ifv_morph = MorphNode.transfigure(ifv.owner, cond_node, True)
                elsev_morph = MorphNode.transfigure(elsev.owner, cond_node, False)
                initial_state[key] = traced(MorphNode.combined(ifv_morph, elsev_morph))
            else:
                raise NotImplementedError(
                    "Malevich does not yet support merging of the states "
                    "of the variables that are not traced."
                )


def exec_flow(
    stmts: list[ast.stmt],
    state,
    filename: str | None = None,
    function_name: str | None = None,
    should_be_true: list[str] | None = None,
    should_be_false: list[str] | None = None,
):
    if should_be_true is None:
        should_be_true = []
    if should_be_false is None:
        should_be_false = []

    return_map = list()

    for stmt in stmts:
        if isinstance(stmt, ast.Return):
            value = ast.Expression(body=stmt.value)
            return_value = eval(
                compile(value, '<string>', 'eval'), state[0], state[1]
            )

            return_map.append(({
                **{s: True for s in should_be_true},
                **{s: False for s in should_be_false}
            } or None, return_value))

            return return_value, state, return_map

        elif isinstance(stmt, ast.If):
            if_expr = ast.Expression(body=stmt.test)

            if_expr_value = eval(
                compile(if_expr, '<string>', 'eval'),
                state[0],
                state[1]
            )

            if not isinstance(if_expr_value, traced) or not isinstance(
                if_expr_value.owner, OperationNode
            ):
                raise SyntaxError(
                    "Condition expression within the flow can only "
                    "be a @condition functions and their boolean combinations. "
                    f"The expression evaluated to `{type(if_expr_value).__name__}`.",
                    (filename, 0, stmt.col_offset, 'if ' + ast.unparse(stmt.test))
                )

            tracers = {*autoflow.get_trace_refs(Flow.flow_ref())}

            b_tree = Flow.flow_ref()
            with Flow() as if_tree:
                # Changing all the traced objects in state before
                # the branch to point to the new tree
                autoflow.retrace(b_tree, if_tree)

                # Mark all future traced objects to be conditioned
                # with node mapper. Also, clone all previous node mappers
                # (for nested conditions)
                mapper_id = 'if' + if_expr_value.owner.uuid

                if_tree.clone_node_mappers(b_tree)
                if_tree.register_node_mapper(
                    AddPositiveCondition(if_expr_value.owner.uuid),
                    key=mapper_id
                )
                # -------------------------------------------------------

                cstate = copy_state(state)

                # Recursively execute the branch
                if_return, if_state, if_map = exec_flow(
                    stmt.body,
                    cstate,
                    filename=filename,
                    should_be_true=[*should_be_true, if_expr_value.owner],
                    should_be_false=should_be_false
                )

                if_tree = Flow.flow_ref()


            if stmt.orelse:
                with Flow() as else_tree:
                    for tracer in tracers:
                        tracer.retrace_self(else_tree)

                    mapper_id = 'else' + if_expr_value.owner.uuid
                    else_tree.clone_node_mappers(b_tree)
                    else_tree.register_node_mapper(
                        AddNegativeCondition(if_expr_value.owner.uuid),
                        key=mapper_id
                    )

                    cstate = copy_state(state)

                    else_return, else_state, else_map = exec_flow(
                        stmt.orelse,
                        cstate,
                        filename=filename,
                        should_be_true=should_be_true,
                        should_be_false=[*should_be_false, if_expr_value.owner]
                    )

                    else_tree = Flow.flow_ref()
            else:
                else_return, else_state, else_tree = None, None, None

            full_tree = ExecutionTree() # Merge all the trees

            for edge in [
                *if_tree.tree,
                *(else_tree.tree if else_tree else []),
                *Flow.flow_ref().tree
            ]:
                # NOTE: List-merge is safe here, but in future
                # it should be encapsulated in special method
                full_tree.put_edge(*edge)

            # Safety first: retrace all the tracers to the new tree
            # and hardset the tree in the flow
            autoflow.retrace(Flow.flow_ref(), full_tree)
            autoflow.retrace(if_tree, full_tree)
            if else_tree:
                autoflow.retrace(else_tree, full_tree)

            # TODO: Hardset is error-prone, need to find a better way
            # with weak refs probably
            Flow._hardset_tree(full_tree)

            if if_return is not None:
                Flow.flow_ref().register_node_mapper(
                    AddNegativeCondition(if_expr_value.owner.uuid)
                )
                return_map.extend(if_map)

                if else_return is None:
                    should_be_false.append(if_expr_value.owner)

            if else_return is not None:
                Flow.flow_ref().register_node_mapper(
                    AddPositiveCondition(if_expr_value.owner.uuid)
                )
                return_map.extend(else_map)

                if if_return:
                    should_be_true.append(if_expr_value.owner)


            # merge_detached_tree_inplace(Flow.flow_ref(), if_tree)
            # if else_tree:
            #     merge_detached_tree_inplace(Flow.flow_ref(), else_tree)

            # state = retrace(Flow.flow_ref(), state)
            combine_branch_states_inplace(state, if_state, else_state, if_expr_value)
        else:
            exec(compile(ast.Module(body=[stmt], type_ignores=[
        ]), '<string>', 'exec'), state[0], state[1])

    return None, state, return_map

class synthetic_tuple(tuple):  # noqa: N801
    def __getattr__(self, item) -> NoReturn:
        if item == '__name__' or item == '__make_tuple':
            return super().__getattribute__(item)
        raise AttributeError(
            "You cannot get an attribute of a result. "
            "Results of function calls are only supposed to be passed as "
            "positional arguments to other functions."
        )

    def __call__(self, *args, **kwargs) -> NoReturn:
        raise TypeError(
            f"You cannot call a result {self.__name__}. "
            "Results of function calls are only supposed to be passed as "
            "positional arguments to other functions."
        )

    def __getitem__(self, item) -> NoReturn:
        raise TypeError(
            "You cannot get an item from a result. "
            "Results of function calls are only supposed to be passed as "
            "positional arguments to other functions. Do not use `result[item]`"
        )

    def __iter__(self) -> Iterator:
        if not self.__make_tuple:
            raise TypeError(
                "You cannot iterate over or unpack a result. "
                "Results of function calls are only supposed to be passed as "
                "positional arguments to other functions."
            )
        return super().__iter__()

    def tuple(self) -> tuple:
        self.__make_tuple = True
        _t = tuple(self)
        self.__make_tuple = False
        return _t

def exec_synthetic_flow(
    stmts: list[ast.stmt],
    state,
    filename: str | None = None,
    function_name: str | None = None,
    should_be_true: list[str] | None = None,
    should_be_false: list[str] | None = None,
):
    collection_args = {}
    document_nodes = {}
    targets = {}

    def synthetic_table(*args, **kwargs):
        return ('__table__', args, kwargs)

    def synthetic_collection(*args, **kwargs):
        uid = uuid.uuid4().hex
        collection_args[uid] = (args, kwargs)
        return synthetic_tuple(('__collection__', uid,))

    def synthetic_document(*args, **kwargs):
        uid = uuid.uuid4().hex
        document_nodes[uid] = (args, kwargs)
        return synthetic_tuple(('document', uid,))

    processor_nodes = {}
    def create_synthetic_processor(name: str):
        def synthetic_processor(*args, **kwargs):
            uid = uuid.uuid4().hex
            node = (name, uid)
            for i, a in enumerate(args):
                if isinstance(a, synthetic_tuple):
                    a = a.tuple()
                Flow.flow_ref().put_edge(a, node, i)
            processor_nodes[uid] = (name, args, kwargs)
            return synthetic_tuple((name, uid,))
        return synthetic_processor

    globals_, locals_ = state

    locals_['collection'] = synthetic_collection
    locals_['document'] = synthetic_document
    locals_['table'] = synthetic_table


    class Synth(ast.NodeVisitor):
        def visit_Call(self, node):
            if isinstance(node.func, ast.Name):
                if node.func.id not in locals_:
                    locals_[node.func.id] = create_synthetic_processor(node.func.id)
            self.generic_visit(node)

    synth = Synth()
    return_value = None

    with Flow() as flow:
        for stmt in stmts:
            synth.visit(stmt)
            if isinstance(stmt, ast.Return):
                value = compile(ast.Expression(body=stmt.value), '<string>', 'eval')
                return_value = eval(value, globals_, locals_)
            elif isinstance(stmt, ast.Assign):
                if not isinstance(stmt.value, ast.Call):
                    raise RuntimeError(
                        "You are not allowed to assign to variables "
                        "to non-call expressions."
                    )
                call_value = eval(compile(ast.Expression(body=stmt.value), '<string>', 'eval'), globals_, locals_)
                assert isinstance(call_value, synthetic_tuple)

                for target in stmt.targets:
                    if isinstance(target, ast.Name):
                        if target.id not in targets:
                            targets[target.id] = [call_value]
                        else:
                            targets[target.id].append(call_value.tuple())
                exec(f"{stmt.targets[0].id} = {call_value!r}", globals_, locals_)

            elif isinstance(stmt, (ast.If, ast.While)):
                raise RuntimeError(
                    "You are not allowed to use 'if' or 'while' statements."
                )
            else:
                try:
                    code = ast.unparse(stmt)
                    exec(compile(ast.Module(body=[stmt], type_ignores=[]), 'flow.py', 'exec'), globals_, locals_)
                except Exception as e:
                    # Create a new exception with the desired information
                    raise RuntimeError(
                        f"Could not execute the code in the flow.\n"
                        f">>> {type(e).__name__}: {e.args[0]} in line\n"
                        f">>> {code}\n"
                    ) from None



    return flow, return_value, (collection_args, document_nodes, processor_nodes), targets






def boot_flow(
    function: Callable,
    __globals: dict,
    __locals: dict,
    /,
    *function_args,
    **function_kwargs,
):
    arguments = inspect.signature(function).parameters

    for key, p in enumerate(arguments.values()):
        if key not in __locals:
            __locals[key] = p.default

    # Map positional only
    for i, param in enumerate(arguments.values()):
        if param.kind == param.POSITIONAL_ONLY:
            __locals[param.name] = function_args[i]

    # Map rest of the positional arguments
    for i, param in enumerate(arguments.values()):
        if param.kind == param.POSITIONAL_OR_KEYWORD and i < len(function_args):
            __locals[param.name] = function_args[i]

    # Map varargs
    varargs_ = ()
    name = None
    for i, param in enumerate(arguments.values()):
        if param.kind == param.VAR_POSITIONAL:
            name = param.name
            varargs_ = function_args[i:]
            break

    if name:
        __locals[name] = varargs_
    # Map keywords
    for key, value in function_kwargs.items():
        if key in __locals:
            __locals[key] = value

    # Check __locals to see if all arguments are mapped
    for key, p in arguments.items():
        if key not in __locals:
            raise TypeError(f"Missing required argument: {key}")

    func_body = inspect.getsource(function)

    indent_ = re.search(
        r'^(?P<INDENT>\s*)def', func_body,
        flags=re.MULTILINE
    ).group('INDENT')

    func_body = re.sub(rf'^{indent_}', '', func_body, flags=re.MULTILINE)

    body = ast.parse(func_body).body[0].body
    return_value, _, return_map = exec_flow(
        body, (__globals, __locals), inspect.getsourcefile(function)
    )

    # return_map.append((None, return_value))

    return return_map



import inspect
import json
from typing import Optional, Type
from uuid import uuid4

import pandas as pd
from malevich_space.schema.schema import SchemaMetadata
from pydantic import BaseModel

import malevich._autoflow.tracer as gn
from malevich._utility import pd_to_json_schema
from malevich.models import Collection, CollectionNode, CollectionOverride
from malevich.table import table


class CollectionNameMeta(type):
    def __getitem__(cls, name_model: tuple[str, Type[BaseModel]]) -> type:

        item, model = name_model
        if model and not inspect.isclass(model):
            raise TypeError(
                "The second argument of collection[] annotation "
                "should be a `pydantic` model"
            )
        if model and not issubclass(model, BaseModel):
            raise TypeError(
                "The second argument of collection[] annotation "
                "should be a `pydantic` model"
            )

        class CollectionProto:
            __malevich_collection_name__: str = item
            __malevich_collection_scheme__: dict = model.model_json_schema() if model else None  # noqa: E501

        return CollectionProto

class collection(metaclass=CollectionNameMeta):  # noqa: N801
    @staticmethod
    def override(
        data: table,
    ) -> CollectionOverride:
        return CollectionOverride(data=data)

    def __new__(
        cls,
        name: str,
        *,
        file: Optional[str] = None,
        df: Optional[pd.DataFrame] = None,
        scheme: SchemaMetadata = None,
        alias: Optional[str] = None,
        persistent: bool = False,
    ) -> gn.traced[CollectionNode]:
        """Creates a collection from a file or a dataframe

        Collection holds tabular data and is subject to be passed to processors.

        Empty collections are allowed and can be used as placeholders
        but will work only when running with Runners.

        .. warning::

            The :func:`collection <malevich._meta.collection>` function can only
            be used inside the :func:`flow <malevich._meta.flow>` function.

        Args:
            name (str, optional): Name of the collection. Defaults to None.
            file (Optional[str], optional): Path to the file. Defaults to None.
            df (Optional[pd.DataFrame], optional): Dataframe. Defaults to None.
            scheme (Optional[SchemaMetadata], optional): Schema of the collection.
            alias (Optional[str], optional):
                An optional short name for the collection.
                If not set, will be infered from the name.
            persistent (bool, optional):
                If the flag is set, the collection is only updated
                when the collection is not present on the platform.
                If persistent collection with the same name is already present,
                the data is not updated. Defaults to False.

        Raises:
            AssertionError: If both file and data are provided

        Returns:
            Collection: a traced object that is subject to be passed to processors
        """
        # Commented in favor of empty collections
        # for runners (Jan 10, 2024, v0.1.2)

        # assert any(
        #     [file is not None, df is not None]
        # ), "Either file or data must be provided"
        # ______________________________________

        if file:
            assert df is None, "Cannot provide both file and data"
            df = pd.read_csv(file)

        name = name or 'auto-collection-' + uuid4().hex[:8]

        # Retrieve JSON Schema
        # from pandas DataFrame
        if df is not None:
            pd_scheme = scheme or json.dumps(pd_to_json_schema(df))
        else:
            pd_scheme = scheme
            df = pd.DataFrame([])

        node = CollectionNode(
            collection=Collection(
                collection_id=name,
                collection_data=df,
                persistent=persistent,
            ),
            scheme=pd_scheme,
        )
        if alias:
            node.alias = alias
        else:
            node.alias = name

        return gn.traced(node)



from functools import cache

import pandas as pd
from malevich_space.ops import SpaceOps
from pandas.core.api import DataFrame as DataFrame

from ..base import BaseResult


class SpaceCollectionResult(BaseResult[list[pd.DataFrame]]):
    """Result obtained running a flow with Space.

    Represents a set of collections obtained after running
    a flow with Space.
    """
    def __init__(
        self,
        run_id: str,
        in_flow_id: str,
        space_ops: SpaceOps
    ) -> None:
        super().__init__()
        self.run_id = run_id
        self.in_flow_id = in_flow_id
        self._ops = space_ops

    def num_elements(self) -> int:
        return len(self.get())

    @cache
    def get(self) -> list[pd.DataFrame]:
        """Fetches a list of DataFrames with the actual result

        Returns:
            list[:class:`DataFrame`]: The actual result
        """
        return [
            pd.DataFrame(x.raw_json)
            for x in self._ops.get_results(
                run_id=self.run_id,
                in_flow_id=self.in_flow_id,
            )
        ]

    @cache
    def get_dfs(self) -> list[DataFrame]:
        """Returns a list of DataFrames from the result.

        Simply calls :meth:`get()`

        Returns:
            list[:class:`DataFrame`]: The list of DataFrames
        """
        return self.get()

    @cache
    def get_df(self) -> DataFrame:
        """Returns a DataFrame from the result if it is a single DataFrame.

        Simply calls :meth:`get()` and returns the first element
        if the list has only one element, otherwise raises an error.

        Returns:
            :class:`DataFrame`: The result DataFrame
        """
        if len(results := self.get()) == 1:
            return results[0]



import rich

from malevich.manifest import ManifestManager


def init():
    try:
        if ManifestManager().local:
            rich.print("\nProject is already initialized.")
            return
    except FileNotFoundError:
        pass

    ManifestManager.create_manifest()

    rich.print(
        "Initialized [purple]Malevich[/purple] project:\n"
        " - [light_black][i]malevich.yaml[/i] contains project settings[/light_black]\n"
        " - [light_black][i]malevich.secrets.yaml[/i] keeps project secrets[/light_black]"   # noqa: E501
    )



from typing import Annotated

import pydantic_yaml as pdyml
import rich
import typer
from malevich_space.schema import Setup
from pydantic import ValidationError

from malevich.manifest import ManifestManager


def init(path_to_setup: Annotated[str, typer.Argument(...)]) -> None:
    try:
        setup = pdyml.parse_yaml_file_as(Setup, path_to_setup)
    except ValidationError as err:
        rich.print(
            f"[b red]Setup file [white]{path_to_setup}[/white] is not a correct configuration[/b red]\n"  # noqa: E501
        )
        for _err in err.errors():
            rich.print(f"\t- {_err['msg']} at [pink]{_err['loc'][0]}[pink]")
        rich.print("\n[red]Could not initialize Space installer[/red]")
    else:
        manf = ManifestManager()

        setup.space.password = manf.put_secret(
            "space_password", setup.space.password)
        ManifestManager().put("space", value=setup.space)

        rich.print(
            "\nMalevich Space configuration [green]successfully[/green] added to the manifest\n"  # noqa: E501
        )



import json
import os
from functools import cache, cached_property
from typing import (
    Any,
    Dict,
    ForwardRef,
    Generic,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

import pandas as pd
from pydantic import BaseModel
from typing_extensions import TypeVarTuple, Unpack

Schemes = TypeVarTuple('Schemes')
Scheme = TypeVar('Scheme')

_docs_first_k_show = 3
DELIMITER = "-" * 50


def _is_M(type: Any) -> bool:  # noqa: N802, ANN401
    return hasattr(type, "__origin__") and type.__origin__.__module__ == __name__ and type.__origin__.__name__ == "M"  # noqa: E501


class M(Generic[Scheme]):
    """Special indicator to be used in :class:`DFS` to denote variable number of inputs.

    See :class:`DFS` for more details.
    """
    pass


class DF(Generic[Scheme], pd.DataFrame):
    """Wrapper class for tabular data.

    DF (and DFS) classes are used to denote tabular data.
    They can specify the scheme of the data. The scheme can be a reference
    tp class adecorated with :func:`malevich.square.jls.scheme`
    or simply a string containing the name of the scheme.

    DF may follow an interface of different data frames implementation.
    But, most of the time, it is just a wrapper of :class:`pandas.DataFrame`.
    So, you can use all the methods of :class:`pandas.DataFrame` directly on DF object.

    .. warning::

        You should not construct DF directly. Instead, when
        returning results from processors, you should use
        any of supported data frames classes directly.

        For example, you may return a :class:`pandas.DataFrame`:

        .. code-block:: python

            import pandas as pd

            @processor()
            def my_processor(df: DF):
                return pd.DataFrame(...)

    """
    def __init__(self, df: Union[pd.DataFrame, Type[BaseModel], Dict, List, 'Doc', 'Docs']) -> None:    # noqa: E501
        if df is None:
            super().__init__(df)
            return

        if isinstance(df, Docs):
            df = df.parse(recursive=True)
        elif isinstance(df, Doc):
            df = [df.dict()]
        elif isinstance(df, Dict):  # Doc
            df = [df]
        elif issubclass(df.__class__, BaseModel):
            df = [df.model_dump()]
        else:
            assert isinstance(df, List) or isinstance(df, pd.DataFrame), f"DF create: expected pd.DataFrame (or Docs, Doc, BaseModel, List, Dict), found {type(df)}"    # noqa: E501
        super().__init__(df)

    def cast(self, scheme: str) -> None:
        raise NotImplementedError("scheme cast not yet implemented")    # TODO

    def scheme(self) -> None:
        raise NotImplementedError("scheme not yet implemented")         # TODO

    @cached_property
    def scheme_name(self) -> Optional[str]:
        """Returns the name of the scheme of the data frame."""
        scheme = self._scheme_cls()
        if scheme is None:
            return None
        if hasattr(scheme, "__name__"):
            return scheme.__name__
        if hasattr(scheme, "_name"):
            return scheme._name
        if isinstance(scheme, str):
            return scheme
        if isinstance(scheme, ForwardRef):
            return scheme.__forward_arg__
        return scheme

    @cached_property
    def _scheme_cls(self) -> Optional[Any]:  # noqa: ANN401
        if hasattr(self, "__orig_class__"):
            return self.__orig_class__.__args__[0]
        return None


class DFS(Generic[Unpack[Schemes]]):
    """Wrapper class for tabular data.

    DFS is a container for multiple DFs. It is used to denote an
    output of processors that return multiple data frames.

    Each of the elements of DFS is also a :class:`DF` or :class:`DFS`.
    """
    def __init__(self) -> None:
        """set dfs with init"""
        self.__dfs: List[Union[DF, DFS, OBJ, Doc, Docs, None]] = []
        self.__inited = False

    def init(self, *dfs: Union[str, pd.DataFrame, Type[BaseModel], Dict, List], nested: bool = False) -> 'DFS': # noqa: E501
        """must be called after __init__, nested should be False"""
        assert not self.__inited, "DFS already inited"
        self.__inited = True
        self.__init(list(dfs), nested)
        return self

    def __add_jdf(self, df: Union[str, pd.DataFrame, Type[BaseModel], Dict, List], type) -> None: # noqa: E501
        if isinstance(df, str):
            self.__dfs.append(OBJ(df))
        elif (hasattr(type, "__origin__") and type.__origin__ is Doc) or type is Doc:
            self.__dfs.append(type(df).init())
        elif (type is Any or type is None) and (isinstance(df, Dict) or issubclass(df.__class__, BaseModel)):   # noqa: E501
            self.__dfs.append(Doc[type](df))
        else:
            self.__dfs.append(DF[type](df))

    def __init(self, dfs: List[Union[str, pd.DataFrame, Type[BaseModel], Dict, List]], nested: bool = False) -> None:   # noqa: E501
        types = self.__orig_class__.__args__ if hasattr(self, "__orig_class__") else [Any for _ in dfs]  # noqa: E501
        many_df_index = None
        for i, type in enumerate(types):
            if _is_M(type):
                assert not nested, "nested M in DFS"
                if many_df_index is not None:
                    raise Exception("more than one M in DFS")
                else:
                    many_df_index = i
        if many_df_index is None:
            assert len(types) == len(dfs), f"wrong arguments size: expected {len(types)}, found {len(dfs)}"  # noqa: E501
            for df, type in zip(dfs, types):
                self.__add_jdf(df, type)
        else:
            assert len(types) - 1 <= len(dfs), f"wrong arguments size: expected at least {len(types) - 1}, found {len(dfs)}"    # noqa: E501
            for df, type in zip(dfs[:many_df_index], types[:many_df_index]):
                self.__add_jdf(df, type)
            count = len(dfs) + 1 - len(types)
            if count != 0:
                type_many = types[many_df_index].__args__[0]
                temp = DFS[tuple([type_many] * count)]().init(*dfs[many_df_index:many_df_index + count], nested=True)  # noqa: E501
                self.__dfs.append(temp)
            else:
                self.__dfs.append(None)
            for df, type in zip(dfs[many_df_index + count:], types[many_df_index + 1:]):
                self.__add_jdf(df, type)

    def __len__(self) -> int:
        """Returns the number of elements in the DFS"""
        return len(self.__dfs)

    def __getitem__(self, key: int) -> Union[DF, 'DFS', 'OBJ', 'Doc', 'Docs', None]:
        """Returns the i-th element of the DFS"""
        return self.__dfs[key]

    def __iter__(self) -> Iterator[Union[DF, 'DFS', 'OBJ', 'Doc', 'Docs', None]]:
        """Returns an iterator over the elements of the DFS"""
        return iter(self.__dfs)

    def __repr__(self) -> str:
        return f"\n{DELIMITER}\n".join(map(lambda x: str(x), self))


class Sink(Generic[Unpack[Schemes]]):
    """Wrapper class to denote a specific type inputs to processor

    Normally, each argument in processor function signature corresponds to
    exactly one output of the previous processor or exactly one collection.

    To denote a processor that is able to accept a variable number of inputs,
    you should use this class.

    Argument of type `Sink` should be the only argument
    of the processor function besides Context.

    ```python

        from typing import Any
        from malevich.square import Sink, DFS, M, processor

        @processor()
        def merge_tables_sink(dfs: Sink["sql_tables"]):
            pass

        @processor()
        def merge_tables_dfs(dfs: DFS[M["sql_tables"]]):
            pass
    ```

    Here, we have two processors. `Sink[schema]` is
    equivalent to `List[DFS[M[schema]]]`.

    The difference between two processors lies in the fact that
    the first one can be connected to any number of processors
    that return data frames with scheme "sql_tables", while the
    second one can be connected to exactly one processor that
    returns any number of data frames with scheme "sql_tables".

    """
    def __init__(self) -> None:
        """set sink with init"""
        self.__data: List[DFS] = []
        self.__inited = False

    def init(self, *list_dfs: List[Union[str, pd.DataFrame, Type[BaseModel], Dict, List]]) -> 'Sink':   # noqa: E501
        """must be called after __init__"""
        assert not self.__inited, "Sink already inited"
        self.__inited = True
        self.__init(list(list_dfs))
        return self

    def __init(self, list_dfs: List[List[Union[str, pd.DataFrame, Type[BaseModel], Dict, List]]]) -> None:  # noqa: E501
        types = self.__orig_class__.__args__ if hasattr(self, "__orig_class__") else None   # noqa: E501
        for dfs in list_dfs:
            cur_types = [Any for _ in dfs] if types is None else types
            self.__data.append(DFS[cur_types]().init(*dfs))

    def __len__(self) -> int:
        return len(self.__data)

    def __getitem__(self, key: int) -> DFS:
        return self.__data[key]

    def __iter__(self) -> Iterator[DFS]:
        return iter(self.__data)

    def __repr__(self) -> str:
        return f"\n{DELIMITER * 2}\n".join(map(lambda x: str(x), self))


class OBJ:
    """Wrapper class that represents files (or folders)

    Used in the same way as :class:`DF`, but provides
    additional functionality to work with files and folders.
    """
    def __init__(self, path: str) -> None:
        self.__path: Any = path

    @property
    def path(self) -> str:
        """A real path to the binary object (or a folder)"""
        return self.__path

    @cached_property
    def raw(self) -> bytes:
        """Reads the binary object from the path"""
        with open(self.__path, 'rb') as f:
            data = f.read()
        return data

    @cached_property
    def as_df(self) -> DF['obj']:
        """df, file paths"""
        paths = []
        if os.path.isfile(self.__path):
            paths.append(self.__path)
        else:
            for address, _, files in os.walk(self.__path):
                for name in files:
                    paths.append(os.path.join(address, name))
        return DF['obj'](pd.DataFrame.from_dict({"path": paths}))

    @cached_property
    def df(self) -> pd.DataFrame:
        """Reads the asset as a data frame as .csv file

        Raises:
            Exception: If asset is not pointed to a .csv file
        """
        return pd.read_csv(self.__path)

    def __repr__(self) -> str:
        return f"OBJ(path={self.__path})"


class Doc(Generic[Scheme]):
    def __init__(self, data: Union[Scheme, Dict, 'Doc', pd.DataFrame, List]) -> None:
        if isinstance(data, List):
            assert len(data) == 1, f"Doc create: too big List, expected size=1, found={len(data)}"  # noqa: E501
            data = data[0]
        if isinstance(data, Doc):
            data = data.parse()
        elif isinstance(data, pd.DataFrame):
            assert data.shape[0] == 1, f"Doc create: too big pd.DataFrame, expected size=1, found={data.shape[0]}"  # noqa: E501
            data = data.to_dict(orient="records")[0]
        assert data is None or isinstance(data, Dict) or issubclass(data.__class__, BaseModel), f"wrong Doc data type: expected Dict, pd.DataFrame or subclass of BaseModel, found {type(data)}"    # noqa: E501
        self.__data: Union[Scheme, Type[BaseModel], Dict] = data

    def parse(self) -> Union[Scheme, Type[BaseModel], Dict]:
        return self.__data

    def init(self) -> 'Doc':
        """must be called after __init__"""
        if self.__data is None:
            return self

        scheme = self._scheme_cls
        if scheme is Any:
            assert isinstance(self.__data, Dict) or issubclass(self.__data.__class__, BaseModel), f"wrong Doc data type: expected Dict or subclass of BaseModel, found {type(self.__data)}" # noqa: E501
            return self
        if scheme is None or scheme.__name__ == "NoneType":
            return self
        if issubclass(scheme, BaseModel):
            if isinstance(self.__data, Dict):
                self.__data = scheme(**self.__data)
            elif isinstance(self.__data, scheme):
                pass
            else:
                self.__data = scheme(**self.__data.model_dump())
        elif isinstance(scheme, str):
            # json_scheme = schemes[scheme]
            raise Exception(f"Doc not yet work with user json schemes: {scheme}")
        else:
            raise Exception(f"Unknown Doc type: {scheme}")
        return self

    def __repr__(self) -> str:
        return f"Doc(__data={{{self.__data}}})"

    @cached_property
    def _scheme_cls(self) -> Optional[Any]: # noqa: ANN401
        if hasattr(self, "__orig_class__"):
            return self.__orig_class__.__args__[0]
        return None

    def dict(self) -> Dict[str, Any]:
        if isinstance(self.__data, Dict):
            return self.__data
        return self.__data.model_dump()

    def json(self) -> str:
        return json.dumps(self.dict())


class Docs(Generic[Scheme]):
    def __init__(self, data: Union[List[Scheme], List[Dict], List[Doc], 'Docs', Scheme, Dict, 'Doc', pd.DataFrame]) -> None:    # noqa: E501
        if data is not None:
            if isinstance(data, pd.DataFrame):
                data = data.to_dict(orient="records")
            if isinstance(data, List):
                if len(data) > 0:
                    first = data[0]
                    assert isinstance(first, Dict) or issubclass(first.__class__, BaseModel) or isinstance(first, Doc),  f"wrong Docs data type: expected List[Dict] or List[Doc] or List[subclass of BaseModel] or Docs, Scheme, Dict, Doc, pd.DataFrame; found List[{type(first)}]"   # noqa: E501
            elif isinstance(data, Docs):
                data = data.parse(recursive=True)
            else:
                assert isinstance(data, Doc) or isinstance(data, Dict) or issubclass(data.__class__, BaseModel), f"wrong Docs data type: expected List[Dict] or List[Doc] or List[subclass of BaseModel] or Docs, Scheme, Dict, Doc, pd.DataFrame; found {type(data)}"  # noqa: E501
        self.__data: List[Doc[Scheme]] = data

    @cache
    def parse(self, *, recursive: bool = False) -> Union[List[Doc[Scheme]], List[Union[Scheme, Type[BaseModel], Dict]]]:    # noqa: E501
        if recursive:
            return [doc.parse() for doc in self.__data]
        return self.__data

    def init(self) -> 'Docs':
        """must be called after __init__"""
        if self.__data is None:
            return self

        scheme = self._scheme_cls
        if isinstance(self.__data, List):
            self.__data = list(map(lambda doc: Doc[scheme](doc).init(), self.__data))
        else:
            self.__data = [Doc[scheme](self.__data).init()]
        return self

    def __len__(self) -> int:
        return len(self.__data)

    def __getitem__(self, key: int) -> Optional[Doc]:
        return self.__data[key]

    def __iter__(self) -> Iterator[Doc]:
        return iter(self.__data)

    def __repr__(self) -> str:
        if len(self.__data) == 0:
            return "Docs(__data=[])"
        if len(self.__data) <= _docs_first_k_show:
            return f"Docs(__data={self.__data[:_docs_first_k_show]}, len={len(self.__data)})"   # noqa: E501
        return f"Docs(__data=[{', '.join(map(str, self.__data[:_docs_first_k_show]))}, ...], len={len(self.__data)})"   # noqa: E501

    @cached_property
    def _scheme_cls(self) -> Optional[Any]: # noqa: ANN401
        if hasattr(self, "__orig_class__"):
            return self.__orig_class__.__args__[0]
        return None

    def json(self) -> str:
        return json.dumps([doc.dict() for doc in self.__data])



import warnings
from typing import Any, Callable, Generic, ParamSpec, TypeVar

from malevich_space.schema import ComponentSchema

Params = ParamSpec("Params")
R = TypeVar("R")

class FlowFunction(Generic[Params, R]):
    """Class-decorator for a flow function

    Captures the function and holds its metadata.
    """
    allowed_flow_function_kwargs = ComponentSchema.model_fields.keys()

    def __init__(
        self,
        f: Callable[Params, R],
        reverse_id: str,
        name: str,
        description: str,
        **kwargs: Any  # noqa: ANN401
    ) -> None:
        """
        Args:
            f (Callable[Params, R]):
                The function that defines the flow
            reverse_id (str):
                Reverse ID of the flow
            name (str):
                Name of the flow
            description (str): Description of the function
            **kwargs (Any):
                Additional arguments to be passed to the flow component.
                See :class:`ComponentSchema` for details.
        """
        self._Captured = f

        _ignored = []
        for k in kwargs:
            if k not in self.allowed_flow_function_kwargs:
                kwargs.pop(k)
                _ignored.append(k)

        if len(_ignored) > 0:
            warnings.warn(
                f"Ignored unexpected arguments to flow function: {', '.join(_ignored)}"
            )

        self.__component = ComponentSchema(
            reverse_id=reverse_id,
            name=name,
            description=description,
            **kwargs
        )

    @property
    def reverse_id(self) -> str:
        return self.__component.reverse_id

    @property
    def name(self) -> str:
        return self.__component.name

    @property
    def description(self) -> str:
        return self.__component.description

    @property
    def component(self) -> str:
        return self.__component

    def __call__(self, *args: Params.args, **kwds: Params.kwargs) -> R:
        return self._Captured(*args, __component=self.__component, **kwds)






from functools import partial, wraps

import malevich_coretools as api

from ..refs import BaseRef, PRSRef
from .service import BaseCoreService


def _get_real_id(s: api.ResultIdsMap, id):
    for k in s.ids:
        if k.id == id:
            return k.realId
    return None

def map_real_pipeline_id(fn):
    @wraps(fn)
    def wrapper(id, *args, **kwargs):
        real_id = None
        return fn(
            id=real_id or (real_id := _get_real_id(api.get_pipelines_map(), id)),
            *args,
            **kwargs
        )
    return wrapper

class PipelineService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

    def id(
        self,
        id: str,
        /,
    ):
        return PRSRef(
            f'PipelineByIdRef({id})',
            create=partial(
                api.create_pipeline,
                pipeline_id=id, auth=self.auth, conn_url=self.conn_url
            ),
            delete=partial(
                map_real_pipeline_id(api.delete_pipeline),
                id, auth=self.auth, conn_url=self.conn_url
            ),
            update=partial(
                map_real_pipeline_id(api.update_pipeline),
                id=id, # mapped to real id
                pipeline_id=id,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            get=partial(
                api.get_pipeline,
                id,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            prepare=partial(
                api.pipeline_prepare,
                pipeline_id=id,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            list=None,
        )


    def all(
        self,
    ):
        return BaseRef(
            'PipelineAllRef',
            create=None,
            delete=partial(
                api.delete_pipelines,
                auth=self.auth, conn_url=self.conn_url
            ),
            update=None,
            get=None,
            list=partial(
                api.get_pipelines,
                auth=self.auth, conn_url=self.conn_url
            ),
        )



from typing import Generic, TypeVar

from .base import BaseRef

CreateFn = TypeVar('CreateFn', bound=callable)
DeleteFn = TypeVar('DeleteFn', bound=callable)
UpdateFn = TypeVar('UpdateFn', bound=callable)
GetFn = TypeVar('GetFn', bound=callable)
PrepareFn = TypeVar('PrepareFn', bound=callable)
RunFn = TypeVar('RunFn', bound=callable)
StopFn = TypeVar('StopFn', bound=callable)
ListFn = TypeVar('ListFn', bound=callable)
ListAllFn = TypeVar('ListAllFn', bound=callable)

class PRSRef(
    Generic[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn, PrepareFn, RunFn, StopFn],
    BaseRef[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn],
):
    def __init__(
        self,
        name: str,
        create: CreateFn | None = None,
        delete: DeleteFn | None = None,
        update: UpdateFn | None = None,
        get: GetFn | None = None,
        list: ListFn | None = None,
        run: RunFn | None = None,
        prepare: PrepareFn | None = None,
        stop: StopFn | None = None,
    ) -> None:
        super().__init__(name, create, delete, update, get, list)
        self.run = run
        self.prepare = prepare
        self.stop = stop




import pickle
import uuid
from enum import Enum
from functools import cache
from threading import Thread
from typing import Callable, Iterable, Optional, Type

import pandas as pd
from gql import gql
from malevich_coretools.abstract.statuses import AppStatus, TaskStatus
from malevich_space.schema import LoadedComponentSchema

from malevich.core_api import BasePlatformSettings
from malevich.models.injections import SpaceInjectable

from ...._autoflow.tracer import traced
from ....types import FlowOutput
from ...nodes.base import BaseNode
from ...nodes.tree import TreeNode
from ...results.space.collection import SpaceCollectionResult
from ..base import BaseTask


class SpaceTaskStage(Enum):
    INTERPRETED     = "interpreted"
    # from Space definition
    GENERATED       = "generated"
    STARTED         = "started"
    STOPPED         = "stopped"



class SpaceTask(BaseTask):
    @staticmethod
    def load(object_bytes) -> 'SpaceTask':
        return pickle.loads(object_bytes)

    @property
    def tree(self) -> TreeNode:
        return self.state.aux.tree

    @property
    def component(self) -> LoadedComponentSchema:
        self.upload()
        return self._component

    @property
    def allow_configurations(self) -> bool:
        return self._component is None

    def __init__(
        self,
        state: 'SpaceInterpreterState',
        component: LoadedComponentSchema | None = None,
        get_component: Callable[..., LoadedComponentSchema] | None = None,
    ) -> None:
        super().__init__()
        assert component is not None or get_component is not None

        self.state = state
        self._component = component
        self.__get_component = get_component
        self._returned = []

    def _deflate(
        self,
        returned: list[traced[BaseNode]],
        alias2infid: dict[str, str]
    ) -> tuple[list[str], list[str]]:
        results_ = []
        infid_ = []
        for x in returned:
            if isinstance(x.owner, TreeNode):
                results_ = x.owner.results
                if not isinstance(results_, list):
                    results_ = [results_]
                returned_ = self._deflate(
                    results_,
                    alias2infid
                )
                results_.extend(returned_[0])
                infid_.extend(returned_[1])
            else:
                results_.append(x.owner.uuid)
                infid_.append(alias2infid[x.owner.alias])
        return results_, infid_

    def prepare(
        self,
        use_v1: bool = False,
        *args,
        **kwargs
    ) -> None:
        if not self.state.aux.flow_id:
            self.state.aux.flow_id = self.component.flow.uid

        if use_v1:
            task_id = self.state.space.build_task(
                flow_id=self.component.flow.uid,
                host_id=self.state.component_manager.host.uid
            )
            self.state.aux.task_id = task_id[0]
        else:
            task_id = self.state.space.build_task_v2(
                flow_id=self.component.flow.uid,
                host_id=self.state.component_manager.host.uid
            )
            self.state.aux.task_id = task_id['uid']
            self.state.aux.core_task_id = task_id['coreId']


        self.state.space.boot_task(
            task_id=self.state.aux.task_id,
        )

    def run(
        self,
        override: dict[str, pd.DataFrame] = {},
        webhook_url: str | None = None,
        *args,
        **kwargs
    ) -> str:
        if not self.state.aux.task_id:
            raise Exception(
                "Attempt to run a task which is not prepared. "
                "Please prepare the task first."
            )

        start_schema = self.state.space.get_task_start_schema(
            self.state.aux.task_id,
        )

        coll_override = {
            **self.state.collection_overrides
        }

        __override = {}
        for alias, df in override.items():
            for inj in self.get_injectables():
                if alias == inj.alias:
                    __override[inj.snapshot_flow_id] = df
        override = __override

        in_flow_ca = {}
        id_to_ca = {}
        overrides = []
        for sch in start_schema:
            try:
                in_flow_ca = self.state.space.get_ca_in_flow(
                    flow_id=self.state.aux.flow_id,
                    in_flow_id=sch.in_flow_id
                )
                id_to_ca[sch.in_flow_id] = in_flow_ca
            except Exception:
                # TODO fix!
                continue

        for in_flow_id, df in override.items():
            uid = self.state.space.create_collection(
                host_id=self.state.host.uid,
                # core_id=f'override-{coll_id}-{state.interpretation_id}',
                core_alias=f'override-{in_flow_id}-{self.state.interpretation_id}-{uuid.uuid4().hex[:4]}',
                # schema_core_id=schema.core_id,
                docs=[
                    row.to_json()
                    for _, row in df.iterrows()
                ]
            )
            coll_override[id_to_ca[in_flow_id]] = uid


        for sch in start_schema:
            try:
                in_flow_ca = id_to_ca[sch.in_flow_id]
                overrides.append({
                    "inFlowCompUid": sch.in_flow_id,
                    "caUid": coll_override[in_flow_ca],
                    "caAlias": sch.injected_alias
                })
            except Exception:
                # TODO fix!
                continue

        self.state.aux.run_id = self.state.space.run_task(
            task_id=self.state.aux.task_id,
            ca_override=overrides,
            webhook=webhook_url,
        )
        return self.state.aux.run_id

    def configure(
        self,
        *operations: str,
        # Configurable parameters
        platform: str = 'base',
        platform_settings: BasePlatformSettings | str = None,
        # Rest of the parameters for compatibility
        **kwargs
    ) -> None:
        if not self.allow_configurations:
            raise Exception(
                "Intepreter already uploaded a new component to Malevich Space, so "
                "configurations are not available. To configure the task, use "
                "`Space(configurable=True)` and call `upload()` afterwards to apply "
                "custom configurations"
            )
        for operation in operations:
            for component in self.state.flow.components:
                if component.alias == operation:
                    component.limits = platform_settings

    def get_interpreted_task(self) -> BaseTask:
        return self

    def get_stage(self) -> SpaceTaskStage:
        if not self.state.aux.task_id:
            return SpaceTaskStage.INTERPRETED

        request = gql(
            """
            query GetTaskBootState($task_id: String!) {
                task(uid: $task_id) {
                    details {
                    bootState
                    }
                }
            }
            """
        )
        response = self.state.space.client.execute(
            request,
            variable_values={'task_id': self.state.aux.task_id}
        )
        state = response['task']['details']['bootState']
        try:
           return SpaceTaskStage[state.upper()]
        except Exception:
            import warnings
            warnings.warn(
                "API returned bootState that "
                "is not in `SpaceTaskStage` enumerator"
            )
            return SpaceTaskStage.INTERPRETED

    def get_stage_class(self) -> Type[SpaceTaskStage]:
        return SpaceTaskStage

    def interpret(self, interpreter=None) -> None:
        return None

    def dump(self) -> bytes:
        return pickle.dumps(self)

    def stop(self, *args, **kwargs) -> None:
        self.state.space.change_task_state(
            task_id=self.state.aux.task_id,
            target_state='stop'
        )
        return

    @cache
    def get_injectables(self) -> list[SpaceInjectable]:
        if self.state.aux.task_id:
            alias_to_snapshot = self.state.space.get_snapshot_components(
                task_id=self.state.aux.task_id
            )
        else:
            alias_to_snapshot = {}

        alias_to_in_flow_id = {
            x.alias: x.uid
            for x in self.component.flow.components
            if x.collection is not None
        }

        alias_to_rev_id = {
            x.alias: x.reverse_id
            for x in self.component.flow.components
            if x.collection is not None
        }

        aliases = set()
        # aliases.update(alias_to_snapshot.keys())
        aliases.update(alias_to_in_flow_id.keys())
        return [
            SpaceInjectable(
                alias=a,
                in_flow_id=alias_to_in_flow_id.get(a, None),
                snapshot_flow_id=alias_to_snapshot.get(a, None),
                reverse_id=alias_to_rev_id.get(a, None)
            )
            for a in aliases
        ]

    def get_operations(self) -> list[str]:
        return [x.alias for x in self.get_injectables()]

    def commit_returned(self, returned: FlowOutput) -> None:
        self._returned = returned

    async def async_results(
        self,
        run_id: Optional[str] = None,
        fetch_timeout: int = 150,
        *args,
        **kwargs
    ) -> Iterable[SpaceCollectionResult]:
        """internal"""
        returned = self._returned
        if returned is None:
            return None

        if isinstance(returned, traced):
            returned = [returned]

        alias2infid = self.state.space.get_snapshot_components(
            run_id=run_id or self.state.aux.run_id
        )

        infid2alias = {
            k: v for v, k in alias2infid.items()
        }

        _, infid_ = self._deflate(
            returned,
            alias2infid
        )

        rs_, cs_ = self.state.space.get_run_status(
            run_id=self.state.aux.run_id
        )

        if rs_ == AppStatus.FAIL:
            raise Exception(
                "Run failed. No results could be fetched..."
            )

        exc_message = None
        if rs_ != AppStatus.COMPLETE.value:
            finished_ = {
                x.in_flow_comp_id for x in cs_ if x.status == AppStatus.COMPLETE.value
            }
            to_be_finished_ = set(infid_)
            async for update in self.state.space.subscribe_to_status(
                run_id or self.state.aux.run_id,
                fetch_timeout
            ):
                if isinstance(update, str):
                    if update == TaskStatus.COMPLETE.value:
                        break
                    if update == TaskStatus.FAIL.value:
                        exc_message = "Run failed. No results could be fetched..."
                        break

                else:
                    if update.status == AppStatus.FAIL.value:
                        exc_message = (
                            f"App {infid2alias[update.in_flow_comp_id]} failed. "
                            "No results could be fetched..."
                        )
                    if update.status == AppStatus.COMPLETE.value:
                        finished_.add(update.in_flow_comp_id)

                    if finished_ == to_be_finished_:
                        break

        if exc_message:
           raise Exception(exc_message)

        return [
            SpaceCollectionResult(
                run_id=run_id or self.state.aux.run_id,
                in_flow_id=i,
                space_ops=self.state.space
            ) for i in infid_
        ]

    def results(
        self,
        run_id: Optional[str] = None,
        fetch_timeout: int = 150,
        *args,
        **kwargs
    ) -> list[SpaceCollectionResult]:
        import asyncio
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="There is no current event loop"
            )
            try:
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    result = None
                    def worker():
                        nonlocal result
                        result = asyncio.run(
                            self.async_results(run_id, fetch_timeout, *args, **kwargs)
                        )

                    thread = Thread(target=worker,)
                    thread.start()
                    thread.join()
                    assert result is not None
                    return result
            except RuntimeError:
                loop = asyncio.new_event_loop()

        def raise_exc(e, *args) -> None:
            raise e

        loop.set_exception_handler(raise_exc)
        return loop.run_until_complete(self.async_results(
            run_id,
            *args,
            **kwargs
        ))

    def upload(self, branch: str | None = None) -> None:
        if not self._component:
            self._component = self.__get_component(branch=branch)




from typing import Any, Callable, Generic, Literal, ParamSpec, TypeVar

from ._deploy import Space
from .models.results import SpaceCollectionResult

ProcFunArgs = ParamSpec("ProcFunArgs")
ProcFunReturn = TypeVar("ProcFunReturn")

class FlowFunctionStub(Generic[ProcFunArgs, ProcFunReturn]):
    def __init__(self, fn, reverse_id) -> None:
        self._fn_ = fn
        self.__call__ = self._fn_call
        self.reverse_id = reverse_id

    def _fn_call(
            self,
            version=None,
            branch=None,
            *,
            run_deployment_id: str | None = None,
            run_task_policy: Literal['only_use', 'use_or_new', 'no_use']='use_or_new',
            get_task: bool = False,
            wait_for_results: bool = True,
            webhook_url: str | None = None,
            **kwargs
    ) -> list[SpaceCollectionResult]:
        task = Space(
            reverse_id=self.reverse_id,
            branch=branch,
            version=version,
            deployment_id=run_deployment_id,
            task_policy=run_task_policy
        )
        if get_task:
            return task

        injectables = task.get_injectables()
        overrides = {}
        for col in injectables:
            data = kwargs.get(col.alias, None)
            if data is not None:
                overrides[col.alias] = data
        run_id = task.run(override=overrides, webhook_url=webhook_url)
        if wait_for_results:
            return task.results()
        return run_id


    __call__ = _fn_call

def installed_flow(reverse_id: str) -> Callable:
    def decorator(fn: Callable) -> FlowFunctionStub[Callable[..., Any], Any]:
        return FlowFunctionStub(fn, reverse_id)
    return decorator



from collections import deque
from typing import Any, Generic, Iterable, Iterator, Optional, TypeVar

T = TypeVar("T")
LinkType = TypeVar("LinkType", bound=Any)


class BadEdgeError(Exception):
    def __init__(self, message: str, edge: tuple[T, T, LinkType]) -> None:
        super().__init__(message)
        self.edge = edge

    def __str__(self) -> str:
        return f"{super().__str__()}: {self.edge}"


class ExecutionTree(Generic[T, LinkType]):
    """An utility class holding a low-level representation of the execution graph

    The class provides a set of basic operations on the graph, such as
    adding edges, pruning, and traversing.

    The tree is stored both as set of nodes and as a list of edges. The nodes
    can be any hashable and comparable class, while the edges are triples
    of the form `(callee, caller, link)` where `caller` and `callee` are
    nodes and `link` is an arbitrary object.

    The connection is directed from `callee` to `caller`. This is
    because :code:`caller(callee)` notation requires the callee to be
    evaluated first.

    .. mermaid::

        graph LR
            Callee --> Caller

    """

    def __init__(self, tree: Optional[list[tuple[T, T, LinkType]]] = None) -> None:
        if tree is not None:
            self.tree = tree
        else:
            self.tree = []
        self.nodes_ = set()
        for u, v, _ in self.tree:
            self.nodes_.add(u)
            self.nodes_.add(v)

        self._node_map  = {}
        self._edge_map  = {}

    def remove_node_mapper(self, key: str) -> None:
        self._node_map.pop(key)

    def remove_edge_mapper(self, key: str) -> None:
        self._edge_map.pop(key)

    def clone_node_mappers(self, other: 'ExecutionTree') -> None:
        self._node_map = {**other._node_map}

    def clone_edge_mappers(self, other: 'ExecutionTree') -> None:
        self._edge_map = {**other._edge_map}

    def register_node_mapper(
        self, mapper: 'BaseNodeMapper', key: str | None = None
    ) -> None:
        """Register a node mapper

        Args:
            mapper (BaseNodeMapper): The node mapper to register
        """
        if key is None:
            key = id(mapper)
        self._node_map[key] = mapper

    def register_edge_mapper(
        self, mapper: 'BaseEdgeMapper', key: str | None = None
    ) -> None:
        """Register an edge mapper

        Args:
            mapper (BaseEdgeMapper): The edge mapper to register
        """
        if key is None:
            key = id(mapper)
        self._edge_map[key] = mapper

    def add_node(self, node: T, from_node: bool | None = None) -> None:
        """Add a node to the execution tree

        Args:
            node (T): The node to add
        """
        for mapper in self._node_map.values():
            mapper(node, from_node=from_node)

        if node not in self.nodes_:
            self.nodes_.add(node)


    def put_edge(self, callee: T, caller: T, link: LinkType = None) -> None:
        """Add an edge to the execution tree

        Args:
            callee (T): The source node (a.k.a. Callee)
            caller (T): The target node (a.k.a. caller)
            link (LinkType, optional):
                An arbitrary object that represents the edge payload
                (e.g. argument name). Defaults to None.

        Raises:
            BadEdgeError: If the edge already exists, or if the edge is a self-edge
        """
        if any(x[0] == callee and x[1] == caller and x[2] == link for x in self.tree):
            raise BadEdgeError("Edge already exists", (callee, caller, link))
        if any(x[0] == caller and x[1] == callee and x[2] == link for x in self.tree):
            raise BadEdgeError("Edge already exists", (callee, caller, link))
        if caller == callee:
            raise BadEdgeError("Self-edge", (callee, caller, link))

        self.add_node(callee, from_node=True)
        self.add_node(caller, from_node=False)

        for mapper in self._edge_map.values():
            mapper(callee, caller, link)
        self.tree.append((callee, caller, link))

    def prune(self, outer_nodes: list[T]) -> None:
        """Removes specified nodes from the execution tree

        Args:
            outer_nodes (list[T]): List of nodes to remove
        """
        self.tree = [
            x for x in self.tree
            if x[0] not in outer_nodes
        ]

        self.nodes_ = set([
            x for x in self.nodes_
            if x not in outer_nodes
        ])

    def roots(self) -> Iterable[tuple[int, T]]:
        return [
            (i, x) for i, x in enumerate(self.tree)
            if not any(y[1] == x[0] for y in self.tree)
        ]

    def edges_from(self, node: T) -> None:
        """Returns all edges starting from the specified node"""
        return [n for n in self.tree if n[0] == node]

    def edges_to(self, node: T) -> None:
        return [n for n in self.tree if n[1] == node]

    def traverse(self) -> Iterator[tuple[T, T, LinkType]]:
        """Traverse the execution tree

        The traversal is performed in a multi-source
        breadth-first manner.

        Returns:
            Generator[T]: Generator of nodes
        """

        # Mark visited nodes
        visited = [False] * len(self.tree)

        # Find roots
        roots = self.roots()

        # Traverse
        q = deque(maxlen=len(self.tree))
        for i, r in roots:
            # BFS
            q.append((i, r,))

        while q:
            j, edge = q.popleft()

            if visited[j]:
                continue

            yield edge
            visited[j] = True

            q.extend(
                filter(
                    lambda x: x[1][0] == edge[1] and not visited[x[0]],
                    enumerate(self.tree)
                )
            )

    def topsort(self) -> Iterator[tuple[T, T, LinkType]]:
        sort_ = []
        s = [r[0] for _, r in self.roots()]
        edges = list(enumerate(self.tree))
        mask = [False] * len(edges)
        while s:
            n = s.pop()
            if n not in sort_:
                sort_.append(n)
            for i, (_, to, _) in filter(
                lambda x: x[1][0] == n and not mask[x[0]], edges
            ):
                mask[i] = True
                in_ = list(filter(lambda x: x[1][1] == to and not mask[x[0]], edges))
                add = True
                for j, (m, _, _) in in_:
                    if m != n and not mask[j]:
                        add = False
                if add:
                    s = [to, *s]

        return sort_


    def leaves(self) -> Iterable[T]:
        """Returns all leaves of the execution tree"""
        return (
            x[1] for x in self.traverse()
            if not any(y[0] == x[1] for y in self.tree)
        )

    @staticmethod
    def connected(a: 'ExecutionTree', b: 'ExecutionTree') -> bool:
        """Checks if two execution trees are connected"""
        return any(
            x[0] == b and x[1] == a for x in a.traverse()
        ) or any(
            x[0] == a and x[1] == b for x in b.traverse()
        )

    def nodes(self) -> Iterable[T]:
        """Returns all nodes of the execution tree"""
        return list(self.nodes_)

    def cast_link_types(self, type) -> None:
        """Cast the link types in the tree"""
        for i, (u, v, link) in enumerate(self.tree):
            self.tree[i] = (u, v, type(link))


from typing import Any, Optional

from pydantic import BaseModel

from .injections import SpaceInjectable


class Dependency(BaseModel):
    package_id: str
    version: Optional[str] = None
    installer: str
    options: Optional[Any] = None

    def simple(self) -> dict[str, Any]:
        return {
            **self.options.model_dump()
        }

    def compatible_with(
        self,
        other: 'Dependency',
        compatability_strategy: 'CompatabilityStrategy' = None
    ) -> bool:
        pass

    def checksum(self) -> str:
        pass


class Integration(BaseModel):
    version: Optional[str] = None
    branch: Optional[str] = None
    deployment: Optional[str] = None
    injectables: list[SpaceInjectable] = []



import os
import sys
from pathlib import Path as _Path

MALEVICH_HOME = os.getenv("MALEVICH_HOME", os.path.expanduser('~/.malevich/'))
MALEVICH_CACHE =  os.path.expanduser(os.getenv("MALEVICH_CACHE", '~/.malevich/cache'))
DB_PATH = '.db'

class Paths:
    @staticmethod
    def home(*path, create: bool = False, create_dir: bool = False) -> str:
        path = list(path)
        if create:
            os.makedirs(os.path.join(MALEVICH_HOME, *path[:-1]), exist_ok=True)
            _Path(os.path.join(MALEVICH_HOME, *path)).touch()
        if create_dir:
            os.makedirs(
                os.path.join(MALEVICH_HOME, *path),
                exist_ok=True
            )
        return os.path.join(MALEVICH_HOME, *path)

    @staticmethod
    def cache(*path) -> str:
        return os.path.join(MALEVICH_CACHE, *path)

    @staticmethod
    def pwd(*path, create: bool = False) -> str:
        if create:
            p = _Path(os.path.join(os.getcwd(), *path))
            p.touch()
            return str(p)
        return os.path.join(os.getcwd(), *path)

    @staticmethod
    def module(*path) -> str:
        return os.path.join(os.path.dirname(sys.modules['malevich'].__file__), *path)

    @staticmethod
    def db() -> str:
        return Paths.home(DB_PATH)

    @staticmethod
    def templates(*args: os.PathLike) -> str:
        return Paths.module('_templates', *args)

    @staticmethod
    def local(*args: os.PathLike, create: bool = False, create_dir: bool = False) -> str:
        return Paths.home('.local', *args, create=create, create_dir=create_dir)

    @staticmethod
    def local_results(
        operation_id: str,
        run_id: str,
        index: int,
        create: bool = False,
        create_dir: bool = False,
    ) -> str | None:
        for path in os.listdir(
            Paths.local(
                'results',
                operation_id,
                run_id,
                create=create,
                create_dir=create_dir
            )
        ):
            if path.startswith(str(index)):
                return Paths.local(
                    'results',
                    operation_id,
                    run_id,
                    path,
                    create=create,
                    create_dir=create_dir
                )
        return None




from collections import defaultdict
import uuid
import weakref
from hashlib import sha256
from typing import Any, Callable, Generic, Optional, TypeVar

from pydantic import BaseModel, Field

from .flow import Flow
from .tree import ExecutionTree


class root(BaseModel):   # noqa: N801
    """A root element of the execution tree

    Used as a default value for the owner of the traced objects
    """

    id: str = Field(..., default_factory=lambda: uuid.uuid4().hex)

    def __eq__(self, other: "root") -> bool:
        if not isinstance(other, root):
            return False
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(sha256(self.id.encode()).hexdigest)


T = TypeVar("T", bound=Any)


class autoflow(Generic[T]):  # noqa: N801
    """Autoflow is a bridge between traced objects and the execution tree

    It holds both the reference to the execution tree and the reference
    to the traced object. Can be reattached to another traced object. Provides
    interfaces for reporting a new dependency in the execution tree.
    """

    _trace_map: dict[weakref.ref[ExecutionTree], set['autoflow']] = defaultdict(set)
    _tracers = 0

    @staticmethod
    def get_tree_ref(tree: ExecutionTree) -> weakref.ref[ExecutionTree] | None:
        for t in autoflow._trace_map.keys():
            if t() is not None and t() is tree:
                return t
        return weakref.ref(tree)

    @staticmethod
    def get_trace_refs(tree: ExecutionTree) -> set['autoflow']:
        tree = autoflow.get_tree_ref(tree)
        if tree is None:
            raise ValueError("Tree not found")
        return autoflow._trace_map[tree]


    @staticmethod
    def retrace(old_tree: ExecutionTree, new_tree: ExecutionTree, /) -> None:
        new_tree_ref = autoflow.get_tree_ref(new_tree)
        for a in autoflow.get_trace_refs(old_tree):
            a._tree_ref = new_tree_ref
            a._notify_spectators(old_tree, new_tree)
            autoflow._trace_map[new_tree_ref].add(a)

    def retrace_self(self, new_tree: ExecutionTree) -> 'autoflow':
        new_tree_ref = autoflow.get_tree_ref(new_tree)

        autoflow._trace_map[self._tree_ref].remove(self)
        self._tree_ref = new_tree_ref
        self._notify_spectators(self._tree_ref(), new_tree)
        autoflow._trace_map[new_tree_ref].add(self)
        return self

    def __init__(self, tree: ExecutionTree[T, Any]) -> None:
        self._tree_ref = autoflow.get_tree_ref(tree)
        self._component_ref = None
        self._trace_map[self._tree_ref].add(self)
        self._tracers += 1
        self._spectators = []

    def attach(self, component: T) -> None:
        """Attach the autoflow to a traced object"""
        self._component_ref = component

    def _notify_spectators(self, old_tree: ExecutionTree, new_tree: ExecutionTree):
        for spectator in self._spectators:
            spectator(self, old_tree, new_tree)

    def spectate_retrace(
        self, callback: Callable[['autoflow', ExecutionTree, ExecutionTree], None]
    ):
        self._spectators.append(callback)

    def calledby(self, caller: 'traced', argument: Optional[str] = None) -> None:
        """Report a new dependency in the execution tree"""
        assert isinstance(caller, traced), "Caller must be a traced object"
        self._tree_ref().put_edge(self._component_ref, caller, argument)

    def __deepcopy__(self, *args):
        copy = super().__deepcopy__(*args)
        autoflow._trace_map[copy._tree_ref].add(copy)
        self._tracers += 1
        return copy

    def __copy__(self, *args):
        copy = super().__copy__(*args)
        autoflow._trace_map[copy._tree_ref].add(copy)
        self._tracers += 1
        return copy



class traced(Generic[T]):  # noqa: N801
    """
    Traced objects are wrappers around regular Python objects
    They hold a reference to the current execution tree and
    provides an access to :class:`autoflow` to enable dependency tracking

    These objects CAN ONLY BE CREATED inside a flow.
    """

    _autoflow: autoflow
    """Autoflow bridge"""

    __keep_tree_on_deepcopy__ = True

    def __init__(
        self,
        owner: T = root()
    ) -> None:
        assert (
            Flow.isinflow()
        ), "Tried to create a tracked variable outside of a flow"
        self._owner = owner
        self._autoflow = autoflow(Flow.flow_ref())
        self._autoflow.attach(self)

    def claim(self, owner: T) -> None:
        """Claim the ownership of the tracer"""
        self._owner = owner

    @property
    def owner(self) -> T:
        if hasattr(self, '_owner'):
            return getattr(self, '_owner')
        else:
            return None

    def __eq__(self, other: "traced") -> bool:
        if not isinstance(other, traced):
            return False
        return self.owner == other.owner

    def __str__(self) -> str:
        if not isinstance(self.owner, traced):
            return f'{self.owner.__str__()}ᵗ'
        else:
            raise RuntimeError(
                "Traced object is not supposed to be nested"
            )

    def __repr__(self) -> str:
        if not isinstance(self.owner, traced):
            return f'traced({self.owner.__repr__()})'
        else:
            raise RuntimeError(
                "Traced object is not supposed to be nested"
            )

    def __hash__(self) -> int:
        return hash(self.owner)

    def __copy__(self, *args):
        return self

    def __deepcopy__(self, *args):
        return self

    def __getitem__(self, *args):
        return tracedLike(self._owner.__getitem__(*args))


class tracedLike(traced[T]):  # noqa: N801
    """This class is used to immitate the traced object without actually tracing it

    It is used to provide homogeneous interface for tree nodes
    without attaching them to a particular tree.

    These objects CAN BE CREATED from outside of a flow.
    """
    def __init__(self, owner: T = root()) -> None:
        self._owner = owner

    @property
    def _autoflow(self) -> autoflow:
        raise ValueError(
            "Tracedlike object does not have an autoflow. Use `traced` instead."
        )

    @property
    def owner(self) -> T:
        return self._owner

    def claim(self, owner: T) -> None:
        return super().claim(owner)

    def __str__(self) -> str:
        return super().__str__() + "*"

    def __repr__(self) -> str:
        return super().__repr__() + "*"




import rich
import typer
from rich.prompt import Prompt
from rich.table import Table

from malevich._utility import flatdict
from malevich.manifest import ManifestManager

app = typer.Typer(name="manifest", help="Manage the manifest file (malevich.yaml)")
secrets = typer.Typer(name="secrets", help="Manage secrets stored in manifest")


DISCLAIMER = """
[bold red]DISCLAIMER[/bold red]
[center]
\tNot manifested internal secrets are candidates for being removed without notice.
\tTo manifest a secret, add it to the [b]'secrets'[/b] section of the [yellow on black]manifest.secrets.yaml[/yellow on black] file.
\tAdditonally, you can use the [yellow on black]malevich secrets fix[/yellow on black] command to interactively manifest secrets.
[/center]
"""  # noqa: E501


@secrets.command("scan", help="Scan the manifest for secrets")
def scan(
    only_number: bool = typer.Option(
        False,
        "--only-number",
        "-n",
        help="Only print the number of secrets found",
    ),
) -> None:
    manf = ManifestManager()
    flat = flatdict(manf.as_dict())
    manifested = manf.get_secrets()['secrets']

    secrets = [x for x in flat if x[1] and ManifestManager.is_secret(x[1])]

    table = Table(title="Secrets", box=rich.box.HORIZONTALS)
    table.add_column("Secret", justify='right')
    table.add_column("Path")
    table.add_column("Manifested", justify='center')

    rich.print(
        f"\n\n[white]Found [spring_green1]{len(secrets)}[/spring_green1] "
        "secrets in manifest[/white]"
    )
    rich.print("\n")
    should_print_legend = False
    for secret in secrets:
        sec = secret[1]
        if only_number:
            sec = sec.split("#")[1]
        tick = "[b green]✔[/b green]" if secret[1] in manifested else "[b red]✘[/b red]"
        if secret[1] in manifested and manifested[secret[1]]['external']:
            tick = '[b yellow]⚠[/b yellow]'
            should_print_legend = True
        table.add_row(sec, ".".join(map(str, secret[0])), tick)
    rich.print(table)
    if should_print_legend:
        rich.print("[bold yellow]⚠[/bold yellow] - external secret")
    rich.print("\n")

    if len(secrets) - len(manifested) > 0:
        rich.print(DISCLAIMER)


@secrets.command(
    "restore", help="Interactively restore the manifest from the secrets file"
)
def fix() -> None:
    manf = ManifestManager()
    flat = flatdict(manf.as_dict())
    secrets = [x for x in flat if x[1] and ManifestManager.is_secret(x[1])]
    manifested = manf.get_secrets()['secrets']
    not_manifested = [x for x in secrets if x[1] not in manifested]

    rich.print(f"\n\n[white]Found [spring_green1]{len(not_manifested)}[/spring_green1] "
               "secrets in manifest[/white]")
    rich.print("\n")

    for secret in not_manifested:
        rich.print(f'\tSecret: {secret[1]}')
        rich.print(f'\tUsed as: {".".join(map(str, secret[0]))}')
        rich.print("\n")
        key = Prompt.ask("Secret key", default=secret[0][-1])
        value = Prompt.ask("Secret value")
        mark_external = Prompt.ask("Mark as external? (y/n)") == "y"
        secret_key = manf.put_secret(key, value, external=mark_external)
        manf.put(*secret[0], value=secret_key, append=False)
        manf.save()
        rich.print(
            f"\n[green]Secret updated: [/green] {secret[1]} -> {secret_key}\n")


@app.command("query", help="Query the path in manifest")
def query(path: list[str], resolve_secrets: bool = False) -> None:
    __q = ManifestManager().query(*path, resolve_secrets=resolve_secrets)
    if __q is None:
        rich.print("[red]Not found[/red]")
    else:
        rich.print(__q)

@app.command("show", help="Show manifest file")
def show() -> None:
    manf = ManifestManager()
    rich.print(manf.path)
    rich.print(manf.as_dict())

app.add_typer(secrets, name="secrets")



from typing import Annotated

import rich
import typer

from malevich._utility.package import package_manager
from malevich.manifest import manf
from malevich.models import Dependency


def list_packages(
    print_installers: Annotated[bool, typer.Option("--installers", "-i")] = False,
    print_versions: Annotated[bool, typer.Option("--versions", "-v")] = False,
) -> None:
    for package in package_manager.get_all_packages():
        try:
            dep_ = Dependency(**manf.query('dependencies', package))
        except Exception:
            continue

        message_ = f"{package}"
        if print_installers or print_versions:
            message_ += " [bright_black]("
        if print_installers:
            message_ += f"installer={dep_.installer}, "
        if print_versions:
            message_ += f"version={dep_.version}  "
        if print_installers or print_versions:
            message_ = message_[:-2] + ")[/bright_black]"

        rich.print(message_)



import json
from collections import defaultdict
from typing import Optional

import malevich_coretools as core
from malevich_coretools import (
    AlternativeArgument,
    Argument,
    Condition,
    JsonImage,
    Processor,
    Result,
)
from malevich_space.schema import ComponentSchema

from malevich._autoflow import traced
from malevich._core import (
    result_collection_name,
)
from malevich._core.service.service import CoreService
from malevich._utility import CacheManager, LogLevel, Registry, cout, unique
from malevich.constants import CORE_INTERPRETER_IN_APP_INFO_KEY, DEFAULT_CORE_HOST
from malevich.interpreter import Interpreter
from malevich.models import (
    Action,
    ArgumentLink,
    AssetNode,
    BaseNode,
    CollectionNode,
    CoreInterpreterState,
    CoreRegistryEntry,
    CoreTask,
    DocumentNode,
    InjectedAppInfo,
    InterpretationError,
    OperationNode,
    TreeNode,
    VerbosityLevel,
)

cache = CacheManager()
registry = Registry()

_levels = [
    LogLevel.Info,
    LogLevel.Warning,
    LogLevel.Error,
    LogLevel.Debug
]

_actions = [
    Action.Interpretation,
    Action.Preparation,
    Action.Run,
    Action.Results
]


def _log(
    message: str,
    level: int = 0,
    action: int = 0,
    step: bool = False,
    *args,
) -> None:
    cout(
        _actions[action],
        message,
        verbosity=VerbosityLevel.AllSteps if step else VerbosityLevel.OnlyStatus,
        level=_levels[level],
        *args,
    )


names_ = defaultdict(lambda: 0)


def _name(base: str) -> int:
    names_[base] += 1
    return names_[base]



class CoreInterpreter(Interpreter[CoreInterpreterState, CoreTask]):
    """Interpret flows to be run on Malevich Core

    Malevich Core is a computational cloud of Malevich. It provides
    low-level access to the computational resources of Malevich. This
    interpreter utilizes its API to run your flows.

    .. note::

        The interpreter can operate with dependencies installed with
        both `image` and `space` installer.

    Prepare
    ---------

    The prepare hook is a simple proxy to the `task_prepare` function
    of Malevich Core API. All the parameters are optional.

    Options:

    *   :code:`with_logs (bool)`:
            If set, return prepare logs if True after end
    *   :code:`debug_mode (bool)`:
            If set, displays additional information about errors
    *   :code:`info_url (str)`:
            URL to which the request is sent. If not specified, the default url is used.
            Rewrite msg_url from configuration if exist
    *   :code:`core_manage (bool)`:
            If set, the requests will be managed by Core, otherwise by the DAG Manager.
    *   :code:`with_show (bool)`:
            Show results (like for each operation, default equals with_logs arg)
    *   :code:`profile_mode (str)`:
            If set, provides more information in logs.
            Possible modes: :code:`no`, :code:`all`, :code:`time`,
            :code:`df_info`, :code:`df_show`

    Run
    -----

    The run hook is a simple proxy to the `task_run` function
    of Malevich Core API. All the parameters are optional.

    Options:

        **Interpreter-specific parameters**

        * :code:`run_id (str)`:
            A custom identifier for the run. If not specified, a random
            identifier is generated.
        * :code:`overrides (dict[str, str])`:
            A dictionary of overrides for the collections. The keys are
            collection names, the values are collection IDs. The option
            is managed by Runners, but can be provided manually.

        **Core-API parameters**

        * :code:`with_logs (bool)`:
            If set return prepare logs if True after end
        * :code:`debug_mode (bool)`:
            If set, displays additional information about errors
        * :code:`with_show (bool)`:
            Show results in logs
        * :code:`profile_mode (str)`:
            If set, provides more information in logs.
            Possible modes: :code:`no`, :code:`all`, :code:`time`,
            :code:`df_info`, :code:`df_show`


    Stop
    ------

    The stop hook is a simple proxy to the `task_stop` function.
    It stops the task and does not have any options.

    Results
    -------

    Results is represented as a list of :class:`malevich.results.core.CoreResult`
    objects. Each object contains a payload which can be an asset, a dataframe or
    a list of such. Use :code:`.get_df()`, :code:`.get_binary()`, :code:`.get_dfs()`
    and :code:`.get_binary_dir()` to extract the results.

    Configure
    ---------

    You can mark an operation to be run on a specific platform. To do so,
    set :code:`platform` and :code:`platform_settings` parameters in the
    :code:`.configure()` method. The :code:`platform` parameter can be
    :code:`base` or :code:`vast`. The :code:`platform_settings` parameter
    is a dictionary of settings for the platform. The settings are
    platform-specific.

    """

    supports_subtrees = False

    def __init__(
        self,
        core_auth: tuple[str, str],
        core_host: str = DEFAULT_CORE_HOST,
    ) -> None:
        super().__init__(CoreInterpreterState())
        # TODO: Remove this (deprecated in favor of service)
        self.__core_host = core_host
        self.__core_auth = core_auth
        self._state.params["core_host"] = core_host
        self._state.params["core_auth"] = core_auth
        # ====================

        self._state.service = CoreService(
            auth=core_auth, conn_url=core_host
        )

        self.update_state()

    def _result_collection_name(self, operation_id: str) -> str:
        return result_collection_name(operation_id)

    def _write_cache(self, object: object, name: str) -> None:
        cache.core.write_entry(
            json.dumps(object),
            entry_name=name,
            entry_group='interpreter/artifacts'
        )

    def _create_cfg_safe(
        self,
        auth: core.AUTH = None,
        conn_url: Optional[str] = None,
        **kwargs,
    ) -> None:
        auth = auth or self.__core_auth
        conn_url = conn_url or self.__core_host
        try:
            cfg_ = core.get_cfg(
                kwargs['cfg_id'],
                auth=auth,
                conn_url=conn_url
            ).id
        except Exception:
            core.create_cfg(
                **kwargs,
                auth=auth,
                conn_url=conn_url
            )
        else:
            core.update_cfg(
                cfg_,
                auth=auth,
                conn_url=conn_url,
                **kwargs
            )

    def interpret(self, node: TreeNode, component: ComponentSchema = None):  # noqa: ANN201
        return super().interpret(node, component)

    def create_node(
        self,
        state: CoreInterpreterState,
        tracer: traced[BaseNode],
    ) -> CoreInterpreterState:
        if isinstance(tracer.owner, OperationNode):
            if tracer.owner.alias is None:
                tracer.owner.alias = unique(tracer.owner.processor_id)
            state.operation_nodes[tracer.owner.alias] = tracer.owner

            extra = registry.get(
                tracer.owner.operation_id,
                model=CoreRegistryEntry
            )

            if not extra.image_ref:
                verbose_ = ""
                if tracer.owner.processor_id is not None:
                    verbose_ += f"processor: {tracer.owner.processor_id}, "
                if tracer.owner.package_id is not None:
                    verbose_ += f"package: {tracer.owner.package_id}."
                if not verbose_:
                    verbose_ = \
                        "The processor and the package cannot be determined " \
                        "(most probably, the app is installed " \
                        "with older version of Malevich)"
                raise InterpretationError(
                    "Found unknown operation. Possibly, you are using "
                    "an outdated version of the environment, or use are not in "
                    "the project directory (with malevich.yaml). \n"
                    + verbose_
                )

            if tracer.owner.is_condition:
                state.conditions[tracer.owner.alias] = Condition(
                    arguments={},
                    cfg=json.dumps({
                        **tracer.owner.config,
                        CORE_INTERPRETER_IN_APP_INFO_KEY: InjectedAppInfo(
                            conn_url=self.__core_host,
                            auth=self.__core_auth,
                            app_id=tracer.owner.alias, # NOTE: was processor_id
                            image_auth=(extra.image_auth_user, extra.image_auth_pass),
                            image_ref=extra.image_ref
                        ).model_dump()
                    }),
                    image=JsonImage(
                        ref=extra.image_ref,
                        user=extra.image_auth_user,
                        token=extra.image_auth_pass
                    ),
                    conditionId=tracer.owner.processor_id,
                )

            else:
                state.processors[tracer.owner.alias] = Processor(
                    arguments={},
                    cfg=json.dumps({
                        **tracer.owner.config,
                        CORE_INTERPRETER_IN_APP_INFO_KEY: InjectedAppInfo(
                            conn_url=self.__core_host,
                            auth=self.__core_auth,
                            app_id=tracer.owner.alias, # NOTE: was processor_id
                            image_auth=(extra.image_auth_user, extra.image_auth_pass),
                            image_ref=extra.image_ref
                        ).model_dump()
                    }),
                    image=JsonImage(
                        ref=extra.image_ref,
                        user=extra.image_auth_user,
                        token=extra.image_auth_pass
                    ),
                    processorId=tracer.owner.processor_id,
                )

            state.results[tracer.owner.alias] = [Result(name=tracer.owner.alias)]

        elif isinstance(tracer.owner, CollectionNode):
            if tracer.owner.alias is None:
                tracer.owner.alias = unique(tracer.owner.collection.collection_id)
            state.collection_nodes[tracer.owner.alias] = tracer.owner
        elif isinstance(tracer.owner, AssetNode):
            if tracer.owner.alias is None:
                tracer.owner.alias = unique(tracer.owner.name)
            state.asset_nodes[tracer.owner.alias] = tracer.owner
        elif isinstance(tracer.owner, TreeNode):
            # Cannot be the case, AbstractInterpreter
            # unwinds the tree is support_subtrees is False
            pass
        elif isinstance(tracer.owner, DocumentNode):
            if tracer.owner.alias is None:
                tracer.owner.alias = unique(tracer.owner.reverse_id)
            state.document_nodes[tracer.owner.alias] = tracer.owner

        _log(f"Node: {tracer.owner.uuid}, {tracer.owner.short_info()}", -1, 0, True)
        return state

    def _add_argument(
        self,
        state: CoreInterpreterState,
        argument: Argument,
        alias: str,
        link: ArgumentLink[BaseNode],
        is_condition: bool = False
    ):
        op_group = state.processors if not is_condition else state.conditions
        a = op_group[alias].arguments.get(link.name, None)
        if a is not None: # has previous arguments
            assert a.alternative or not argument.conditions, (
                "violation: argument with conditions, but no alternative"
                f" {a.model_dump()}"
            )
            if argument.conditions:
                a.alternative.append(argument)
            else:
                assert link.in_sink, (
                    "violation: clash of arguments but no in_sink flag"
                )

                if a.group is None:
                    a.group = []
                a.group.append(argument)
        else:
            if argument.conditions:
                a = AlternativeArgument(alternative=[argument])
            elif link.in_sink:
                a = AlternativeArgument(group=[argument])
            else:
                a = AlternativeArgument(**argument.model_dump())

        op_group[alias].arguments[link.name] = a


    def make_argument(
        self,
        state: CoreInterpreterState,
        node: BaseNode,
        conditions: dict[OperationNode, str] | None = None,
    ) -> Argument | AlternativeArgument:

        conditions = {
            key.alias: value
            for key, value in (conditions or {}).items()
        }

        if isinstance(node, CollectionNode):
            argument = Argument(
                collectionName=state.collection_nodes[node.alias].collection.collection_id,
                conditions=conditions
            )
        elif isinstance(node, OperationNode):
            argument = Argument(
                id=node.alias,
                indices=node.subindex,
                conditions=conditions
            )
        elif isinstance(node, AssetNode):
            argument = Argument(
                collectionName=state.asset_nodes[node.alias].name,
                conditions=conditions
            )
        elif isinstance(node, DocumentNode):
            argument = Argument(
                collectionName=state.document_nodes[node.alias].reverse_id,
                conditions=conditions
            )

        return argument

    def create_dependency(
        self,
        state: CoreInterpreterState,
        from_node: traced[BaseNode],
        to_node: traced[OperationNode],
        link: ArgumentLink[BaseNode],
        conditions: dict[str, bool] | None = None
    ) -> CoreInterpreterState:
        for fcondition, fnode in from_node.owner:
            for tcondition, tnode in to_node.owner:
                if fnode.alias is None:
                    for node in self._state.operation_nodes.values():
                        if node.uuid == fnode.uuid:
                            fnode.alias = node.alias
                            break
                    else:
                        fnode.alias = unique(fnode.uuid)

                cond_stmt = {
                    **(fcondition or {}),
                    **(conditions or {}),
                    **(tcondition or {})
                }
                self._add_argument(
                    state,
                    self.make_argument(state, fnode, cond_stmt),
                    tnode.alias,
                    link,
                    is_condition=(isinstance(tnode, OperationNode) and tnode.is_condition)
                )

                _log(
                    f"Dependency: {fnode.short_info()} -> "
                    f"{tnode.short_info()}, "
                    f"Link: {link.name}"
                    f"", -1, 0, True
                )
        return state

    def before_interpret(self, state: CoreInterpreterState) -> CoreInterpreterState:
        _log("Connection to Core is established.", 0, 0, True)
        _log(f"Core host: {self.__core_host}", 0, 0, True)
        try:
            core.check_auth(
                auth=self.__core_auth,
                conn_url=self.__core_host
            )
        except Exception:
            try:
                core.create_user(
                    auth=self.__core_auth,
                    conn_url=self.__core_host
                )
            except Exception:
                raise Exception(
                    "Cannot connect to Core. Please, check your credentials."
                )
        return state

    def after_interpret(self, state: CoreInterpreterState) -> CoreInterpreterState:
        for operation in state.operation_nodes.values():
            if operation.is_condition:
                continue
            if operation.should_be_true:
                for cond_uid in operation.should_be_true:
                    for condition in state.operation_nodes.values():
                        if condition.uuid == cond_uid:
                            _log(f"Condition {operation.alias}: {condition.alias} == True", -1, step=True)
                            if not state.processors[operation.alias].conditions:
                                state.processors[operation.alias].conditions = {}
                            state.processors[operation.alias].conditions[condition.alias] = True
            if operation.should_be_false:
                for cond_uid in operation.should_be_false:
                    for condition in state.operation_nodes.values():
                        if condition.uuid == cond_uid:
                            if not state.processors[operation.alias].conditions:
                                state.processors[operation.alias].conditions = {}
                            _log(f"Condition {operation.alias}: {condition.alias} == False", -1, step=True)
                            state.processors[operation.alias].conditions[condition.alias] = False

        return state

    def get_task(
        self, state: CoreInterpreterState
    ) -> CoreTask:
        """Creates an instance of task using interpreted state

        Args:
            state (CoreInterpreterV2State): State of the interpreter

        Returns:
            CoreTask: An instance of task
        """
        _log("Task is being compiled...")


        return CoreTask(
            state=state,
            component=self._component,
        )

    def attach(
        self,
        unique_task_hash: str,
        only_fetch: bool = False,
    ) -> CoreTask:
        task = CoreTask(self.state)
        return task.connect(
            unique_task_hash=unique_task_hash,
            only_fetch=only_fetch
        )



import inspect
from functools import wraps
import os
from typing import TypeVar

T = TypeVar('T')

def capture_caller_file(function: T) -> T:
    @wraps(function)
    def wrapped_function(*args, **kwargs):
        f_back_g = inspect.currentframe().f_back.f_globals
        if '__file__' in f_back_g:
            caller_file = f_back_g['__file__']
            if caller_file and caller_file.endswith('.py'):
               from malevich._analytics import manager
               manager.write_artifact({
                   'type': 'file_content',
                   'file': os.path.abspath(caller_file),
                   'content': open(caller_file).read(),
                   'writer': function.__qualname__
               })

        return function(*args, **kwargs)
    return wrapped_function



from functools import partial
from typing import Generic, TypeVar

from ..._utility.core_logging import IgnoreCoreLogs

from ..refs.base import BaseRef

CreateFn = TypeVar('CreateFn', bound=callable)
DeleteFn = TypeVar('DeleteFn', bound=callable)
UpdateFn = TypeVar('UpdateFn', bound=callable)
GetFn = TypeVar('GetFn', bound=callable)
ListFn = TypeVar('ListFn', bound=callable)

T = TypeVar('T')
def wrap_fn(
    fn: T,
    ref: 'BaseRef',
    create: bool = False,
    delete: bool = False,
    update: bool = False,
    get: bool = False,
    list: bool = False,
) -> T:

    def wrapped(*args, **kwargs):
        if create:
            out = fn(*args, **kwargs)
            ref._created = True
            return out
        if delete:
            out = fn(*args, **kwargs)
            ref._deleted = True
            return out
        return fn(*args, **kwargs)
    return wrapped


class ConfigRef(BaseRef[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn], 
              Generic[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn]):

    @property
    def update_or_create(self) -> UpdateFn | None:
        try:
            with IgnoreCoreLogs():
                cfg = self.get()
            assert hasattr(cfg, 'id')
        except Exception:
            return self.create
        else:
            return partial(self.update, id=cfg.id)




import re
from typing import Optional

import rich
import typer
from malevich_space.ops import SpaceOps
from malevich_space.schema import HostSchema, SpaceSetup
from rich.prompt import Prompt

from malevich.constants import DEFAULT_CORE_HOST, PROD_SPACE_API_URL
from malevich.manifest import ManifestManager

from ..prefs import prefs as prefs


def login(
    no_input: bool = typer.Option(False, "--no-input"),
    api_url: str = PROD_SPACE_API_URL,
    core_url: str = DEFAULT_CORE_HOST,
    space_url: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    org_id: Optional[str] = None,
) -> None:
    try:
        if not space_url:
            domain = re.search(r"\/\/(.*)api\.(.+)\/?", api_url)
            left = domain.group(1) if domain.group(1) else ''
            right = '.' + domain.group(2) if domain.group(2) else ''
            space_url = f'https://{left}space{right}/'
            base_space_url = f'https://{left}space{right}'.rstrip('/')
        else:
            domain = re.search(r"\/\/(.*)space\.(.+)\/?", space_url)
            left = domain.group(1) if domain.group(1) else ''
            right = '.' + domain.group(2) if domain.group(2) else ''
            base_space_url = f'{left}space{right}'.rstrip('/')
            api_url = f'https://{left}api{right.rstrip("/")}/'
    except Exception:
        base_space_url = api_url
    if no_input and (username is None or password is None):
        rich.print("[red]You have to set --username and --password parameters, "
                   "if --no-input is used[/red]")
        return False

    manf = ManifestManager()
    rich.print(
        "[b]Welcome to [purple]Malevich Space[/purple]![/b]"
        " The command allows you to connect your account "
        f"to [bright_cyan]{space_url}[/bright_cyan]"
        "[bright_black]\nIf you don't have an account, "
        "please create one and come back.[/bright_black]\n"
    )

    if not username:
        username = Prompt.ask(
            f"E-mail (or Username) on [bright_cyan]{base_space_url}[/bright_cyan]"
        )
    if not password:
        password = Prompt.ask(
            "Password",
            password=True
        )

    if not org_id and not no_input:
        org_id = Prompt.ask(
            "Organization slug (leave blank to use personal account)",
            default=None,
        )

    setup = SpaceSetup(
        api_url=api_url.rstrip('/'),
        username=username,
        password=password,
        org=org_id,
        host=HostSchema(
            conn_url=core_url,
        )
    )

    try:
        SpaceOps(space_setup=setup)
    except Exception:
        rich.print(
            f"\n\n[red]Failed to connect to {space_url}. "
            "Please check your credentials and try again.[/red]"
        )
        return False

    space_password = manf.put_secret("space_password", setup.password)
    setup.password = space_password
    manf.put("space", value=setup)
    rich.print("\nMalevich Space configuration [green]successfully[/green]"
               " added to the manifest\n")

    return True



try:
    from malevich_app.export.secondary.jls_imported import imported3, set_imported3

    if not imported3:
        from malevich_app.jls_lib.utils import *
        from malevich_app.export.jls.jls import input_doc, input_df, input_true, processor, output, condition, scheme, init
        from malevich_app.export.jls.df import M, DF, DFS, Sink, OBJ, Doc, Docs
        from malevich_app.export.defaults.schemes import *
        set_imported3(True)
    from malevich_app.docker_export.schemes import *
except ImportError:
    from .jls import *
    from .utils import *
    from .defaults import *
    from .df import DF, DFS, M, Sink, OBJ, Doc, Docs



from malevich_space.schema import SpaceSetup
from typer import Context, Option, Typer

from malevich._utility import get_core_creds_from_setup
from malevich._utility._try import try_cascade
from malevich.constants import DEFAULT_CORE_HOST
from malevich.core_api import Config, set_host_port, update_core_credentials
from malevich.manifest import manf

from ..space.login import login
from .assets import assets_app
from .endpoints import endpoints_app
from .limits import limits_app

core_app = Typer(name='core')

@core_app.callback()
def callback(
    ctx: Context,
    user: str = Option(None, '--user', '-u', help="Username on Malevich Core"),
    password: str = Option(None, '--password', '-p', help="Password on Malevich Core"),
    host: str = Option(DEFAULT_CORE_HOST, '--host', '-h', help="Host of Malevich Core"),
) -> None:
    def creds():
        nonlocal user, password
        return get_core_creds_from_setup(
            SpaceSetup(**manf.query('space', resolve_secrets=True))
        )

    if user is None or password is None:
        creds_, exc = try_cascade(
            creds,
            lambda *args: login(no_input=False),
        )

        if exc or not creds_:
            raise Exception(
                "Failed to authorize. Run `malevich space login`"
            ) from exc

        user, password = creds_

    ctx.obj = {
        'auth': (user, password),
        'conn_url': host.strip('/') + '/'
    }

    update_core_credentials(user, password)
    set_host_port(host)

@core_app.command(name='whoami')
def whoami(
    show_password: bool = Option(
        False, '--show-password', help='Whether to print passwordof the user'
    )
) -> None:
    """Prints the current user"""
    user, password = Config.CORE_USERNAME, Config.CORE_PASSWORD

    print(f'User: {user}')
    if show_password:
        print(f'Password: {password}')



core_app.add_typer(endpoints_app, name='endpoints', help="Manage endpoints")
core_app.add_typer(assets_app, name='assets', help="Manage binary files")
core_app.add_typer(
    limits_app,
    name='limits',
    help="Retrieve and set limits for cloud resource"
)



from malevich._autoflow import ExecutionTree, traced, tracedLike
from malevich.models import ArgumentLink, BaseNode
from ..models.nodes.morph import MorphNode
from malevich.types import FlowTree

MAX_DEFLAT_DEPTH = 1024
def deflat_edges(
    link: ArgumentLink,
) -> list[tuple[ArgumentLink, traced[BaseNode]]]:
    from malevich.models import TreeNode
    edges_ = link.compressed_edges
    def _done(nodes_: tuple[ArgumentLink, traced[BaseNode]]) -> bool:
        return all([not isinstance(n.owner, TreeNode) for _, n in nodes_])

    i = 0
    while not _done(edges_) and i < MAX_DEFLAT_DEPTH:
        new_edges = []
        for link, node in edges_:
            if isinstance(node.owner, TreeNode) and link.compressed_edges is not None:
                new_edges.extend(link.compressed_edges)
            else:
                new_edges.append((link, node))
        edges_ = new_edges
        i += 1

    if i == MAX_DEFLAT_DEPTH:
        raise RecursionError(
            "Maximum recursion depth reached while deflating edges"
        )

    return edges_

def unwrap_tree(
    tree: FlowTree,
) -> FlowTree:
    """Merges all subtrees into one tree

    Args
        tree (FlowTree): Tree of any depth

    Returns:
        FlowTree: Unwrapped tree
    """
    from malevich.models import TreeNode

    if len(tree.nodes_) == 1:
        return tree

    edges = []
    unwraped = {}
    for edge in tree.traverse():
        u = edge[0].owner
        v = edge[1].owner
        link = edge[2]
        if isinstance(v, TreeNode):
            if not unwraped.get(v.uuid, False):
                for inner_edge in unwrap_tree(v.tree).traverse():
                    if inner_edge[0] == link.shadow_collection:
                        continue
                    edges.append(inner_edge)
                unwraped[v.uuid] = True

            for bridge in deflat_edges(edge[2]):
                f = (
                    tracedLike(u.underlying_node)
                    if isinstance(u, TreeNode)
                    else edge[0]
                )

                edges.append(
                    (f, bridge[1], bridge[0])
                )

        elif isinstance(u, TreeNode):
            edges.append(
                (tracedLike(u.underlying_node), edge[1], edge[2]),)
        else:
            edges.append(edge)

    w_out_mophed_trees = []
    for u, v, link in edges:
        if isinstance(u.owner, MorphNode):
            members = []
            for conds, node in u.owner:
                if isinstance(node, TreeNode):
                    node = node.underlying_node
                members.append((conds, node))
            u = tracedLike(MorphNode(members=members))
            u.owner.correct_self()
        if isinstance(v.owner, MorphNode):
            members = []
            for conds, node in v.owner:
                if isinstance(node, TreeNode):
                    node = node.underlying_node
                members.append((conds, node))
            v = tracedLike(MorphNode(members=members))
            v.owner.correct_self()
        w_out_mophed_trees.append((u, v, link))

    return ExecutionTree(w_out_mophed_trees)



from functools import partial

import malevich_coretools as api

from ..refs import BaseRef, PRSRef
from .base import BaseCoreService


class RunService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

    @property
    def active(self):
        return BaseRef(
            'ActiveRunsRef',
            list=partial(
                api.get_run_active_runs,
                auth=self.auth,
                conn_url=self.conn_url
            ),
        )

    def operation_id(self, operation_id: str):
        return PRSRef(
            'OperationIdRef',
            run=partial(
                api.task_run,
                auth=self.auth,
                conn_url=self.conn_url,
                operation_id=operation_id
            ),
            stop=partial(
                api.task_stop,
                auth=self.auth,
                conn_url=self.conn_url,
                operation_id=operation_id
            ),
        )



import hashlib
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel

from malevich.models.dependency import Dependency


class LocalOptions(BaseModel):
    import_path: Path
    dependencies: Optional[list[str]] = None


class LocalDependency(Dependency):
    installer: str = 'local'
    options: Optional[LocalOptions] = None

    def simple(self) -> dict[str, Any]:
        simpler = {**self.options.model_dump()}
        return {k: v for k, v in simpler.items() if v is not None}

    def checksum(self) -> str:
        return hashlib.sha256(self.model_dump_json().encode()).hexdigest()

    def compatible_with(
        self,
        other: 'Dependency',
        compatability_strategy,
    ) -> bool:
        return False



import os
import tempfile
import zipfile
from typing import Optional

import humanfriendly
import rich
from rich.progress import Progress
from rich.tree import Tree
from typer import Typer

from ..._utility.asset_full_tree import get_asset_full_tree
from malevich.core_api import (
    delete_collection_object,
    get_collection_object,
    get_collection_objects,
    update_collection_object,
)

from ..misc.make import wrap_command

assets_app = Typer(name='assets')


@assets_app.command(name='list')
@wrap_command(get_collection_objects)
def list_assets(
    sizes: bool = False,
    plain: bool = False,
    **kwargs
) -> None:
    kwargs.pop('path', None)
    file_dirs = get_asset_full_tree(
        base_path='/',
        conn_url=kwargs.get('conn_url'),
        auth=kwargs.get('auth'),
    )

    if plain:
        if sizes:
            length_ = max(map(len, file_dirs.files.keys()))
            contents = '\n'.join([
                f'{p.ljust(1 + length_)}\t{humanfriendly.format_size(size)}'
                for p, size in file_dirs.files.items()
            ])
        else:
            contents = '\n'.join(file_dirs.files.keys())
        print(contents)
        return

    tree = Tree('assets')
    for file, size in file_dirs.files.items():
        path = file.split('/')
        node = tree
        for p in path[:-1]:
            node = node.add(p)
        p = path[-1]
        if sizes:
            node.add(
                f'{p} [bright_black]({humanfriendly.format_size(size)})[/bright_black]'
            )
        else:
            node.add(p)
    rich.print()
    rich.print(tree)
    rich.print()


@assets_app.command(name='get')
@wrap_command(get_collection_object)
def get(save: Optional[str] = None, **kwargs) -> None:
    path = os.path.basename(kwargs.get('path'))
    if save:
        bytes_ = get_collection_object(**kwargs)
        if os.path.isdir(save):
            path = os.path.join(save, path)
        with open(path, 'wb') as f:
            f.write(bytes_)
    else:
        get_collection_objects(recursive=False, **kwargs)
        name, size = next(
            get_collection_objects(recursive=False, **kwargs).files.items()
        )
        rich.print(
            f'{name}\t[bright_black]{humanfriendly.format_size(size)}[/bright_black]')


@assets_app.command(name='upload')
@wrap_command(update_collection_object, exclude=['data', 'zip'])
def update(local_path: str, **kwargs) -> None:
    local_path = os.path.abspath(os.path.expanduser(local_path))
    if not os.path.exists(local_path):
        rich.print("\nPath is invalid")
        exit(-1)

    if os.path.isdir(local_path):

        with Progress() as progress:
            files = set().union(*[
                [os.path.join(r, f) for f in fs]
                for r, _, fs in os.walk(local_path)
            ])
            archive_task = progress.add_task(
                'Compressing files',
                total=len(files)
            )
            with tempfile.TemporaryFile('wb+') as tf, zipfile.ZipFile(tf, 'w') as f:
                for fl in files:
                    f.write(fl)
                    progress.update(archive_task, advance=1)
                f.close()
                tf.seek(0)
                bytes_ = tf.read()

                update_collection_object(
                    data=bytes_,
                    zip=True,
                    **kwargs,
                )
    else:
        update_collection_object(
            data=open(local_path, 'rb'),
            zip=False,
            **kwargs,
        )

@assets_app.command(name='delete')
@wrap_command(delete_collection_object)
def delete(**kwargs) -> None:
    delete_collection_object(**kwargs)




from .asset import AssetFactory
from .collection import collection
from .flow import flow
from .run import run
from .document import document

__all__ = [
    "AssetFactory",
    "collection",
    "flow",
    "run",
    "document",
]



import logging

from malevich.core_api import set_logger

logger = logging.getLogger('malevich.core')
handles = [*logger.handlers]
set_logger(logger)

class IgnoreCoreLogs:
    def __enter__(self, *args):
        logger.setLevel(logging.CRITICAL + 1)
        logger.handlers = []
        logger.propagate = False
        set_logger(logger)

        return self

    def __exit__(self, *args) -> None:
        logger.setLevel(logging.NOTSET)
        logger.handlers = handles
        logger.propagate = True
        set_logger(logger)




from enum import Enum

from pydantic import BaseModel

from .actions import Action


class VerbosityLevel(Enum):
    Quiet = 0
    OnlyStatus = 1
    AllSteps = 2


class LogFormat(Enum):
    Plain = "PLAIN"
    Rich = "RICH"


class UserPreferences(BaseModel):
    verbosity: dict[str, int] = {
        Action.Interpretation.value: VerbosityLevel.Quiet.value,
        Action.Preparation.value: VerbosityLevel.OnlyStatus.value,
        Action.Run.value: VerbosityLevel.OnlyStatus.value,
        Action.Stop.value: VerbosityLevel.Quiet.value,
        Action.Installation.value: VerbosityLevel.Quiet.value,
        Action.Removal.value: VerbosityLevel.Quiet.value,
        Action.Results.value: VerbosityLevel.OnlyStatus.value,
        Action.Attachment.value: VerbosityLevel.OnlyStatus.value,
    }
    log_format: LogFormat = LogFormat.Rich
    log_level: str = "INFO"



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar, get_args, overload

import pandas as pd
from pydantic import BaseModel

RealResultType = TypeVar("RealResultType")
DocumentType = TypeVar("DocumentType", bound=BaseModel)

class BaseResult(ABC, Generic[RealResultType]):
    """Result obtained running a flow.

    Encapsulates the result object obtained after running a flow.
    The object can be a DataFrame, a list of DataFrames, binary data, etc.
    """

    def result_type(self) -> Type:
        return get_args(self.__orig_class__)[0]

    @property
    @abstractmethod
    def num_elements(self) -> int:
        """The number of elements (assets/collections) in the result"""
        pass

    @abstractmethod
    def get(self) -> RealResultType:
        """Fetches a real result object (download, etc.)

        Returns:
            RealResultType: The actual result
        """
        pass

    def get_df(self) -> pd.DataFrame:
        """Returns a DataFrame from the result if possible

        Returns:
            pd.DataFrame: The DataFrame
        """
        raise NotImplementedError(
            f"Result {self.__class__.__name__} does not support "
            "conversion to DataFrame"
        )

    def get_dfs(self) -> list[pd.DataFrame]:
        """Returns a list of DataFrames from the result if possible

        Returns:
            list[pd.DataFrame]: The list of DataFrames
        """
        raise NotImplementedError(
            f"Result {self.__class__.__name__} does not support "
            "conversion to list of DataFrames"
        )

    def get_binary(self) -> bytes:
        """Returns binary data from the result if possible

        Returns:
            bytes: The binary data
        """
        raise NotImplementedError(
            f"Result {self.__class__.__name__} does not support "
            "conversion to binary data"
        )

    def get_binary_dir(self) -> dict[str, bytes]:
        """Returns a dict of binary data from the result if possible

        Returns:
            dict[str, bytes]: The dict of binary data
        """
        raise NotImplementedError(
            f"Result {self.__class__.__name__} does not support "
            "conversion to dict of binary data"
        )


    @overload
    def get_document(self, index: int = 0) -> dict:
        """Returns dumped document from the result if possible

        Returns:
            dict: The dumped document
        """
        pass

    @overload
    def get_document(self, index:int = 0, *, model: Type[DocumentType]) -> DocumentType:
        """Returns dumped document from the result if possible

        Returns:
            dict: The dumped document
        """
        pass

    def get_document(
        self,
        index: int = 0,
        *,
        model: Type[DocumentType] | None = None
    ) -> DocumentType:
        """Returns a document from the result if possible

        Returns:
            DocumentType: The document
        """
        pass



from functools import partial, wraps
from typing import TypeVar

import malevich_coretools as api

from ..._utility.core_logging import IgnoreCoreLogs
from ..refs import BaseRef, CollectionRef
from .base import BaseCoreService

T = TypeVar("T")

def map_name_to_id(fn: T, pass_name: bool = False) -> T:
    def wrapper(name, *args, **kwargs):
        with IgnoreCoreLogs():
            if pass_name:
                return fn(
                    id=api.get_collection_by_name(
                        name=name,
                        auth=kwargs["auth"],
                        conn_url=kwargs["conn_url"]
                    ).id,
                    name=name,
                    *args,
                    **kwargs
                )

            return fn(
                id=api.get_collection_by_name(
                    name=name,
                    auth=kwargs["auth"],
                    conn_url=kwargs["conn_url"]
                ).id,
                *args,
                **kwargs
            )
    return wrapper

def update_collection_from_df(id, data, *args, **kwargs):
    docs = [
        api.create_doc(
            data=x,
            name=None,
            auth=kwargs["auth"],
            conn_url=kwargs["conn_url"]
        )
        for x in data.to_dict(orient="records")
    ]
    return api.update_collection(
        id=id,
        ids=docs,
        *args,
        **kwargs
    )

class CollectionService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

    def id(
        self,
        id: str,
        /,
    ):
        return CollectionRef(
            name=f'CollectionByIdRef({id})',
            create=None,
            delete=partial(
                api.delete_collection,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            update=partial(
                update_collection_from_df,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            get=partial(
                api.get_collection,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            list=None,
            get_table=partial(
                api.get_collection_to_df,
                id, auth=self.auth, conn_url=self.conn_url
            ),
        )


    def name(
        self,
        name: str,
        /,
    ):
        return CollectionRef(
            name=f'CollectionByNameRef({name})',
            create=partial(
                api.create_collection_from_df,
                name=name, auth=self.auth, conn_url=self.conn_url
            ),
            update=partial(
                map_name_to_id(update_collection_from_df, pass_name=True),
                name=name, auth=self.auth, conn_url=self.conn_url
            ),
            delete=partial(
                map_name_to_id(api.delete_collection),
                name=name, auth=self.auth, conn_url=self.conn_url
            ),
            get=partial(
                api.get_collection_by_name,
                name, auth=self.auth, conn_url=self.conn_url
            ),
            get_table=partial(
                api.get_collection_by_name_to_df,
                name, auth=self.auth, conn_url=self.conn_url
            ),
            list=partial(
                api.get_collections_by_name,
                name, auth=self.auth, conn_url=self.conn_url
            ),
        )

    def group_name(
        self,
        group_name: str,
        /,
    ):
        return BaseRef(
            name=f'CollectionsByGroupRef({group_name})',
            create=None,
            delete=None,
            update=None,
            get=None,
            list=partial(
                api.get_collections_by_group_name,
                group_name,
                auth=self.auth,
                conn_url=self.conn_url
            )
        )


    def all(self):
        return BaseRef(
            name='AllCollectionsRef',
            create=None,
            delete=partial(
                api.delete_collections,
                auth=self.auth, conn_url=self.conn_url
            ),
            update=None,
            get=None,
            list=partial(
                api.get_collections,
                auth=self.auth, conn_url=self.conn_url
            )
        )






from __future__ import annotations

from typing import Iterator

from pydantic import BaseModel, ConfigDict

from malevich._autoflow.tracer import traced

from .base import BaseNode
from .operation import OperationNode


class MorphNode(BaseNode):
    members: list[tuple[dict[OperationNode, bool] | None, BaseNode]] = []

    def __iter__(self) -> Iterator[tuple[dict[OperationNode, bool] | None, BaseNode]]:
        return iter(self.members)

    @staticmethod
    def transfigure(
        node: BaseNode,
        condition: traced[OperationNode],
        value: bool
    ) -> MorphNode:
        members = []
        for conds, node in node:
            assert not isinstance(node, MorphNode), (
                "violation: MorphNode cannot be nested"
            )
            if conds is None:
                conds = {condition.owner: value}
            else:
                conds[condition.owner] = value
            members.append((conds, node))
        return MorphNode(members=members)


    @staticmethod
    def combined(
        alpha: BaseNode,
        beta: BaseNode
    ) -> MorphNode:
        members = []
        for conds, node in alpha:
            members.append((conds, node))
        for conds, node in beta:
            members.append((conds, node))
        return MorphNode(members=members)


    def correct_self(self) -> None:
        members = []
        for conds, node in self.members:
            if isinstance(node, MorphNode):
                node.correct_self()
                for ncond, nnode in node:
                    if ncond is None:
                        ncond = conds
                    else:
                        ncond.update(conds or {})
                    members.append((ncond, nnode))
            else:
                members.append((conds, node))
        self.members = members



from ..._autoflow.link import AutoflowLink
from ..nodes.base import BaseNode
from .base import BaseEdgeMapper


class LoopEdgeMapper(BaseEdgeMapper):
    def __init__(self, loop_nodes: list[BaseNode]):
        self._ln = loop_nodes

    def __call__(self, source, target, link) -> None:
        if source in self._ln:
            if isinstance(link, AutoflowLink):
                link.loop_argument = True
            else:
                raise ValueError('Invalid link in mapper')
        return True



import json
from typing import Iterable, Optional

from pydantic import Field

from .base import BaseNode


class OperationNodeSpawner:
    RUN_AWAY_ITERATOR_LIMIT = 32

    def __init__(self, base_operation: 'OperationNode') -> None:
        self._index = -1
        self._base = base_operation

    def __next__(self) -> 'OperationNode':
        if self._index > self.RUN_AWAY_ITERATOR_LIMIT:
            raise RuntimeError(
                "Trying to unpack operation result with "
                "exhausitve iterator (like list or for-loop). "
                "Such syntax is not supported, use definite unpacking: "
                f"first, second = {self._base.processor_id}(...)"
            )
        self._index += 1
        return self._base.model_copy(update={'subindex': self._index}, deep=False)


class OperationNode(BaseNode):
    operation_id: str
    processor_id: Optional[str] = None
    package_id: Optional[str] = None
    config: dict = {}
    subindex: list[int] | None = None
    is_condition: bool = False

    should_be_true: list[str] = []
    should_be_false: list[str] = []

    def __eq__(self, other) -> bool:
        if isinstance(other, OperationNode):
            return all([
                self.uuid == other.uuid,
                self.subindex == other.subindex,
                self.should_be_true == other.should_be_true,
                self.should_be_false == other.should_be_false,
            ])
        else:
            return False

    def get_senstivite_fields(self) -> dict[str, str]:
        return {"config": json.dumps(self.config)}

    def set_sensitive_fields(self, values: dict[str, str]) -> None:
        config = values.get("config")
        if config is None:
            return
        self.config = json.loads(config)

    def short_info(self) -> str:
        p_ = ""
        if self.package_id:
            p_ += f"{self.package_id}"
        if self.processor_id:
            p_ += f"::{self.processor_id}"
        return f"{self.__class__.__name__}({self.uuid[:6]}, {p_}, {self.alias})"

    def __hash__(self) -> int:
        return super().__hash__()

    def __getitem__(self, index: int | slice) -> 'OperationNode':
        if isinstance(index, slice):
            assert index.step is None, "Step is not supported"
            start = index.start or 0
            if index.stop is None or index.stop < 0:
                raise ValueError(
                    "Only positive stop index is supported for operation result "
                    "slicing. But either stop index is negative or not provided."
                )
            copy_ = self.model_copy(
                update={'subindex': [i for i in range(start, index.stop)]},
                deep=False
            )
        copy_ = self.model_copy(update={'subindex': [index]}, deep=False)
        return copy_



from .base import CIOps
from .functions import CIFunctions

try:
    from .github import GithubCIOps, DockerRegistry
except Exception:
    import warnings
    warnings.warn(
        "Could not import GitHub operations submodule. Check whether `git` "
        "executable is available in your system."
    )



from malevich_space.schema import LoadedComponentSchema

from malevich._utility import Stub
from malevich.core_api import AppFunctionsInfo
from malevich.manifest import manf
from malevich.models.dependency import Dependency
from malevich.path import Paths

from .installer import Installer
from .space import SpaceDependency, SpaceDependencyOptions


class IsolatedSpaceInstaller(Installer):
    def install(
        self,
        package_id: str,
        component: LoadedComponentSchema,
        app_info: AppFunctionsInfo
    ) -> None:

        if component.app:
            iauth, itoken = manf.put_secrets(
                image_auth_user=component.app.container_user,
                image_auth_password=component.app.container_token,
            )
        else:
            iauth, itoken = None, None

        dependency = SpaceDependency(
            package_id=package_id,
            options=SpaceDependencyOptions(
                reverse_id=component.reverse_id,
                image_ref=component.app.container_ref,
                image_auth_user=iauth,
                image_auth_pass=itoken,
            )
        )

        Stub.from_app_info(
            app_info=app_info,
            path=Paths.module(package_id),
            package_name=package_id,
            dependency=dependency,
            operation_ids={
                op.core_id: op.uid
                for op in component.app.ops
            },
            registry_records={
                str(op.core_id): {
                    "operation_id": op.uid,
                    "reverse_id": component.reverse_id,
                    "branch": component.branch.uid,
                    "version": component.version.uid,
                    "processor_name": op.core_id,
                    "processor_id":  op.core_id,
                    "image_ref": ('dependencies', package_id, 'options', 'image_ref'),
                    "image_auth_user": (
                        'dependencies', package_id, 'options', 'image_auth_user'
                    ),
                    "image_auth_pass": (
                        'dependencies', package_id, 'options', 'image_auth_pass'
                    )
                }
                for op in component.app.ops
            },
            description=component.description
        )

    def restore(self, dependency: Dependency) -> Dependency:
        return None

    def construct_dependency(self, object: dict) -> Dependency:
        return None



from .registry import Registry
from .get_space_leaves import get_space_leaves
from .upload_zip_asset import upload_zip_asset
from .schema import pd_to_json_schema, generate_empty_df_from_schema
from .tree import unwrap_tree, deflat_edges
from .space import (
    get_auto_ops,
    get_core_creds_from_setup,
    resolve_setup
)
from .args import parse_kv_args
from .dicts import flatdict
from .host import fix_host
from .logging import LogLevel, cout
from .tree_node_hash import get_tree_node_hash
from .cache.manager import CacheManager, CacheController
from .core_logging import IgnoreCoreLogs
from .flow_stub import FlowStub
from .package import PackageManager, package_manager
from .stub import Stub, StubFunction, StubIndex, StubSchema
from .unique import unique




from typing import Any, Optional
import uuid

from malevich.models import CoreInjectable
import malevich_coretools as core
from malevich_app import LocalRunner, LocalRunStruct, cfg_translate, pipeline_translate
from pydantic import BaseModel

from malevich._utility.logging import cout
from malevich.models.actions import Action
from malevich.models.nodes.asset import AssetNode
from malevich.models.nodes.base import BaseNode
from malevich.models.nodes.collection import CollectionNode
from malevich.models.nodes.morph import MorphNode
from malevich.models.nodes.tree import TreeNode
from malevich.models.overrides import DocumentOverride, Override
from malevich.models.results.local.infs import LocalResult
from malevich.models.state.local import LocalInterpreterState
from malevich.models.task.interpreted.core import CoreTask
from malevich.models.overrides import DocumentOverride, CollectionOverride, AssetOverride
from malevich.models.collection import Collection
from malevich.path import Paths


class LocalTaskState(LocalInterpreterState):
    config: core.Cfg | None = None
    config_id: str | None = None
    pipeline_id: str | None = None
    unique_task_hash: str | None = None
    operation_id: str | None = None

class LocalTask(CoreTask):
    def __init__(self, state: LocalTaskState) -> None:
        self._returned = None
        self.state = LocalTaskState()

        for key in LocalInterpreterState.model_fields.keys():
            setattr(self.state, key, getattr(state, key))

        self.run_id = None
        self.component = None
        self.__struct = LocalRunStruct(
            import_dirs=list(self.state.import_paths),
            results_dir=Paths.local('results', create_dir=True),
            mount_path=Paths.local('mnt', create_dir=True),
            mount_path_obj=Paths.local('mnt_obj', create_dir=True),
        )
        self.runner = LocalRunner(
            local_settings=self.__struct,
        )

    def get_active_tasks(self) -> None:
        raise RuntimeError(
            "Cannot get active tasks in local mode"
        )

    def get_stage(self) -> None:
        raise RuntimeError(
            "Cannot get stage in local mode"
        )

    def configure(*args, **kwargs) -> None:
        raise RuntimeError(
            "Cannot configure in local mode"
        )

    async def prepare(
        self,
        *args,
        **kwargs
    ) -> tuple[str, str]:
        for node in self.state.asset_nodes.values():
            if node.real_path is None or node.core_path is None:
                raise ValueError(
                    "Real and remote paths must be specified for all assets"
                    " when using local mode."
                )

            self.runner.storage.obj(
                obj_path=node.core_path,
                path_from=node.real_path,
            )

        for node in self.state.document_nodes.values():
            node.core_id = self.runner.storage.data(
                data=node.document.model_dump() if isinstance(node.document, BaseModel) else node.document,
                id=node.core_id,
            )

        for node in self.state.collection_nodes.values():
            node.collection.core_id = self.runner.storage.data(
                data=node.collection.collection_data,
                id=node.collection.core_id,
            )

        if not self.state.config:
            config = core.Cfg(
                collections={
                    v.collection.collection_id: v.collection.core_id
                    for k, v in self.state.collection_nodes.items()
                }
            )

            for an in self.state.asset_nodes.values():
                config.collections = {
                    **config.collections,
                    an.name: an.get_core_path(),
                }

            for dn in self.state.document_nodes.values():
                config.collections = {
                    **config.collections,
                    dn.reverse_id: dn.get_core_path(),
                }

            self.state.config = config

        self.state.operation_id = await self.runner.prepare(
            pipeline_translate(self.get_pipeline(with_hash=False)),
            cfg_translate(self.state.config),
            **kwargs
        )
        
    def _prepare_collection_overrides(
        self,
        injectables: list[CoreInjectable],
        overrides: dict[str, CollectionOverride]
    ) -> dict[str, str]:

        return {
            key: self.runner.storage.data(ref.data)
            for key, ref in overrides.items()
        }


    def _prepare_document_overrides(
        self,
        injectables: list[CoreInjectable],
        overrides: dict[str, DocumentOverride]
    ) -> dict[str, str]:
        return {
            key: '#' + self.runner.storage.data(ref.data.model_dump() if isinstance(ref.data, BaseModel) else ref.data)
            for key, ref in overrides.items()
        }

    async def run(
        self,
        override: dict[str, Override] | None = None,
        config_extension: dict[str, dict[str, Any] | BaseModel] | None = None,
        run_id: Optional[str] = None,
        *args,
        **kwargs
    ) -> None:
        self.run_id = run_id or uuid.uuid4().hex

        if override:
            real_overrides = self._compile_overrides(override)
        else:
            real_overrides = {}

        if config_extension:
            print(config_extension, self._validate_extension(config_extension))
            app_cfg_extensions = self._validate_extension(config_extension)
        else:
            app_cfg_extensions = {}

        if real_overrides or app_cfg_extensions:
            new_config = self.state.config.model_copy(deep=True)
            new_config.collections = {
                **self.state.config.collections,
                **real_overrides,
            }

            new_config.app_cfg_extension = app_cfg_extensions
            print(new_config.app_cfg_extension)
            await self.runner.run(
                *args,
                run_id=self.run_id,
                operation_id=self.state.operation_id,
                cfg=cfg_translate(new_config),
                **kwargs
            )
        else:
            await self.runner.run(
                *args,
                run_id=self.run_id,
                operation_id=self.state.operation_id,
                cfg=cfg_translate(self.state.config),
                **kwargs
            )

        return self.run_id

    async def stop(self, *args, **kwargs) -> None:
        await self.runner.stop(
            *args,
            run_id=self.run_id,
            operation_id=self.state.operation_id,
            **kwargs
        )

    async def results(
        self,
        run_id: Optional[str] = None,
        *args,
        **kwargs
    ) -> list[LocalResult]:
        cout(
            message=f"Scanning results. {Paths.local('results', self.state.operation_id, self.run_id)}",  # noqa: E501
            action=Action.Results
        )
        if self.state.operation_id is None:
            raise Exception("Attempt to run a task which is not prepared. "
                            "Please, run `.prepare()` first.")

        returned = self._returned
        if not self._returned:
            return None

        demorphed_returned = []
        for i in range(len(returned)):
            if isinstance(returned[i][1].owner, MorphNode):
                for morph_conditions, node in returned[i][1].owner.members:
                    demorphed_returned.append(({
                        **(returned[i][0] or {}),
                        **(morph_conditions or {})
                        }, node,
                    ))
            else:
                demorphed_returned.append(returned[i])

        returned = demorphed_returned

        if not run_id:
            run_id = self.run_id

        no_conditions_return = None
        final_result = None

        condition_map = {}
        for condition, return_map in returned:
            if condition is None:
                no_conditions_return = return_map

        if not condition_map:
            final_result = no_conditions_return

        returned = final_result

        if not isinstance(returned, list):
            returned = [returned]

        def _deflat(li: list[BaseNode]) -> list[BaseNode]:
            temp_returned = []
            for r in li:
                if isinstance(r, TreeNode):
                    temp_returned.extend(
                        _deflat(r.results)
                    )
                else:
                    temp_returned.append(r)
            return temp_returned

        returned = _deflat(returned)
        results = [
            LocalResult(
                operation_id=self.state.operation_id,
                run_id=self.run_id,
                name=node.owner.alias,
                login=self.__struct.login,
            )
            for node in returned
        ]

        return results



from functools import partial

import malevich_coretools as api

from ..refs import BaseRef, ConfigRef
from .service import BaseCoreService


class ConfigService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

    def id(
        self,
        id: str,
        /,
    ):
        return BaseRef(
            name=f'ConfigByIdRef({id})',
            create=None,
            delete=partial(
                api.delete_cfg,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            update=partial(
                api.update_cfg,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            get=partial(
                api.get_cfg_real,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            list=None,
        )

    def name(
        self,
        name: str,
        /,
    ):
        return ConfigRef(
            name=f'ConfigByNameRef({name})',
            create=partial(
                api.create_cfg,
                cfg_id=name, auth=self.auth, conn_url=self.conn_url
            ),
            delete=None,
            update=partial(
                api.update_cfg,
                cfg_id=name,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            get=partial(
                api.get_cfg,
                name,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            list=None,
        )

    def all(
        self,
    ):
        return BaseRef(
            name='ConfigAllRef',
            create=None,
            delete=partial(
                api.delete_cfgs,
                auth=self.auth, conn_url=self.conn_url
            ),
            update=None,
            list=partial(
                api.get_cfgs,
                auth=self.auth, conn_url=self.conn_url
            ),
            get=None,
        )



import os

_backend = os.environ.get("MALEVICH_DF_BACKEND", "pandas")

match _backend:
    case "pandas":
        try:
            import pandas as pd
            _table_cls = pd.DataFrame
        except ImportError:
            pass
    case _:
        _table_cls = None

if not _table_cls:
    raise ImportError(
        "No backend found for meta table"
    )

class table(_table_cls):  # noqa: N801
    pass

__all__ = ["table"]



import os
import shutil
import tempfile

import yaml
from malevich_space.constants import ACTIVE_SETUP_PATH
from malevich_space.ops.env import get_file_path

from ...manifest import ManifestManager


class __ManifestAsEnv:  # noqa: N801
    def __init__(self) -> None:
        self.space = None
        self.tmp = None
        self.__path = get_file_path(ACTIVE_SETUP_PATH)

    def __enter__(self) -> None:
        manf = ManifestManager()
        space = manf.query('space', resolve_secrets=True)
        self.space = {'space': space}
        self.tmp = tempfile.NamedTemporaryFile('w+')
        if os.path.exists(self.__path):
            shutil.copyfileobj(
                open(self.__path),
                self.tmp
            )

        self.tmp.seek(0)
        with open(self.__path, 'w') as f:
            yaml.dump(self.space, f)

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        old_space = yaml.load(self.tmp, Loader=yaml.FullLoader)
        self.tmp.close()
        with open(self.__path, 'w') as f:
            yaml.dump(old_space, f)

        if exc_type:
            return False
        return True

manifest_as_env = __ManifestAsEnv()



import malevich_coretools as core

from malevich._utility import fix_host


def scan_core(
    core_host: str,
    image_ref: str,
    image_auth: tuple[str, str],
    core_auth: tuple[str, str] | None = None,
) -> core.abstract.AppFunctionsInfo:
    """Scans Core for image info

    Args:
        core_auth: Core credentials
        core_host: Core host
        image_ref: Docker image reference
        image_auth: Docker image credentials
    """
    try:
        info = core.get_image_info(
            image_ref,
            image_auth=image_auth,
            conn_url=fix_host(core_host),
            parse=True
        )
    except Exception as err:
        raise Exception(
            f"Can't get info for image {image_ref}. Check that "
            "image exists and credentials are correct.\n"
            f"Error: {err}"
        )
    else:
        return info



from collections import defaultdict
from enum import Enum
from typing import Any
from uuid import uuid4

from malevich_space.ops.component_manager import ComponentManager
from malevich_space.ops.space import SpaceOps
from malevich_space.schema.component import ComponentSchema
from malevich_space.schema.flow import (
    FlowSchema,
    InFlowComponentSchema,
    InFlowDependency,
    OpSchema,
)
from malevich_space.schema.host import LoadedHostSchema

from malevich.models import TreeNode


class SpaceAuxParams:
    task_id: str | None
    run_id: str | None
    core_task_id: str | None
    flow_id: str | None
    tree: TreeNode | None

    def __init__(self, **kwargs) -> None:
        self.task_id = kwargs.get('task_id', None)
        self.run_id = kwargs.get('run_id', None)
        self.core_task_id = kwargs.get('core_task_id', None)
        self.flow_id = kwargs.get('flow_id', None)
        self.tree = kwargs.get('tree', None)

    def __getitem__(self, key: str) -> Any:  # noqa: ANN401
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        setattr(self, key, value)

    def __contains__(self, key: str) -> bool:
        return hasattr(self, str(key))

class SpaceInterpreterState:
    """State of the Space interpreter."""

    def __init__(self) -> None:

        # A manager to operate with components
        self.component_manager: ComponentManager = None
        # Flow to be interpreted
        self.flow: FlowSchema = FlowSchema()
        # Space Operations (just for convenience, same as in component_manager)
        self.space: SpaceOps = None
        # Host to run the task
        self.host: LoadedHostSchema = None
        # In flow components``
        self.components: dict[str, InFlowComponentSchema | ComponentSchema] = {}
        # In flow components aliases
        # NOTE: deprecated in favor of node.alias
        self.components_alias: dict[str, str] = {}
        # In flow components configs
        self.components_config: dict[str, dict[str, str]] = {}
        # A mappping from node uuid to node type (collection or operation)
        self.node_to_operation: dict[str, str] = {}
        # A mapping from node uuid to selected operation in the operation component
        self.selected_operation: dict[str, OpSchema] = {}
        # A mapping from collection uid to CA uid for override (if new data is provided)
        self.collection_overrides: dict[str, str] = {}
        # A mapping from node uuid to dependencies
        self.dependencies: dict[str, list[InFlowDependency]] = defaultdict(list)
        # Unique id for the interpretation
        self.interpretation_id: str = uuid4().hex  # as in API interpreter xD
        # A dictionary for storing auxiliary information
        # task_id, flow_id, etc.
        self.aux: SpaceAuxParams = SpaceAuxParams()
        self.children_states: dict[str, SpaceInterpreterState] = {}

    def copy(self) -> 'SpaceInterpreterState':
        state = SpaceInterpreterState()
        state.component_manager = self.component_manager
        state.flow = self.flow
        state.components = self.components
        state.components_alias = self.components_alias
        state.components_config = self.components_config
        state.dependencies = self.dependencies
        state.aux = self.aux
        state.space = self.space
        state.node_to_operation = self.node_to_operation
        state.selected_operation = self.selected_operation
        state.interpretation_id = self.interpretation_id
        state.collection_overrides = self.collection_overrides
        state.host = self.host
        state.children_states = self.children_states
        return state



import atexit
import base64
import logging
import os
from datetime import datetime
from typing import Any

import dill

from malevich._db import Write

from .._db.schema.artifacts import Session
from .utils import decide_session

logger = logging.getLogger('malevich.artifacts')

class ArtifactsManager:
    session_id = decide_session()

    @staticmethod
    def write_artifact(payload) -> None:
        from .._db.schema.artifacts import Artifact

        with Write() as db_ref:

            artifact = Artifact(
                payload=base64.b64encode(dill.dumps(payload)).decode(),
                captured_at=datetime.now(),
                session_id=ArtifactsManager.session_id
            )
            db_ref.add(artifact)
            db_ref.commit()
            pld_str = str(payload)
            if len(pld_str) > 40:
                pld_str = pld_str[:20] + '...' + pld_str[-17:]

            logger.debug(
                f'New artifact: id={artifact.id}, session={ArtifactsManager.session_id}, content={pld_str}'
            )

    @staticmethod
    def finalize_session() -> None:
        import subprocess

        import malevich

        script_path = os.path.join(os.path.dirname(malevich.__file__), '_scripts', 'finalize.py')
        subprocess.run(['python', script_path, str(ArtifactsManager.session_id)])

    @staticmethod
    def get_artifact(id: int) -> dict[str, Any]:
        from .._db.schema.artifacts import Artifact

        with Write() as db_ref:
            artifact = db_ref.query(Artifact).get(id)
            return dill.loads(base64.b64decode(artifact.payload))

    @staticmethod
    def get_all_artifacts() -> list[dict[str, Any]]:
        from .._db.schema.artifacts import Artifact

        with Write(timeout=2) as db_ref:
            artifacts = db_ref.query(Artifact).all()
            return [
                dill.loads(base64.b64decode(artifact.payload))
                for artifact in artifacts
            ]

    @staticmethod
    def get_session_artifacts(session_id: int) -> list[dict[str, Any]]:
        from .._db.schema.artifacts import Artifact

        with Write() as db_ref:
            artifacts = db_ref.query(Artifact).filter(
                Artifact.session_id == session_id
            ).all()
            return [
                dill.loads(base64.b64decode(artifact.payload))
                for artifact in artifacts
            ]

    @staticmethod
    def get_grouped_artifacts() -> dict[int, dict[str, Any]]:
        from .._db.schema.artifacts import Artifact

        with Write() as db_ref:
            artifacts = db_ref.query(Artifact).all()
            grouped = {}
            for artifact in artifacts:
                if artifact.session_id not in grouped:
                    grouped[artifact.session_id] = []
                grouped[artifact.session_id].append(artifact)
            return grouped



if not os.getenv('_MALEVICH_NO_FINALIZE'):
    atexit.register(ArtifactsManager.finalize_session)



from enum import Enum


class Action(Enum):
    Interpretation = "Intrepretation"
    Attachment = "Attachment"
    Preparation = "Preparation"
    Run = "Run"
    Stop = "Stop"
    Installation = "Installation"
    Removal = "Removal"
    Results = "Results"



from typing import Any, Optional

from pydantic import BaseModel


class CoreRegistryEntry(BaseModel):
    image_ref: tuple[str, ...] | str
    processor_id: str
    image_auth_user: Optional[tuple[str, ...] | str] = None
    image_auth_pass: Optional[tuple[str, ...] | str] = None

    def __getattribute__(self, __name: str) -> str:
        from malevich.manifest import manf

        attr_ = super().__getattribute__(__name)
        if __name in ['image_auth_user', 'image_auth_pass', 'image_ref']:
            if isinstance(attr_, tuple):
                return manf.query(*attr_, resolve_secrets=True)
        return attr_

    def __getitem__(self, key: str) -> Any:  # noqa: ANN401
        # Backward compatability with dictionary
        # bypassing `self.__getattribute__`
        # to not query for manifested values
        # in case of access like `extra["image_ref"]`
        return super().__getattribute__(key)



import os
import tempfile
import zipfile
from functools import partial

import malevich_coretools as api

from ..refs import AssetRef
from .service import BaseCoreService


class AssetService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

    @staticmethod
    def _upload(
        path: str,
        file: str | None = None,
        files: list[str] | None = None,
        folder: str | None = None,
        data: bytes | None = None,
        zip: None = None,
        *args,
        **kwargs
    ) -> api.Alias.Info:
        if file is not None:
            with open(file, "rb") as f:
                data = f.read()
        if folder is not None:
            files = files or []
            for root, _, fs in os.walk(folder):
               files.extend(
                   os.path.join(root, f) for f in fs
                )
        if files is not None:
            temp_file = tempfile.NamedTemporaryFile(mode="+rb", delete=False)
            with zipfile.ZipFile(temp_file, "w") as zip_file:
                for f in files:
                    zip_file.write(f, f)
            temp_file.seek(0)
            data = temp_file.read()

        return api.update_collection_object(
            path,
            data=data,
            zip=files is not None,
            *args,
            **kwargs
        )

    def path(
        self,
        path: str,
        /
    ):
        return AssetRef(
            name=f'AssetRef{path}',
            create=partial(
                AssetService._upload,
                path=path,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            update=partial(
                AssetService._upload,
                path=path,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            delete=partial(
                api.delete_collection_object,
                path, auth=self.auth, conn_url=self.conn_url
            ),
            get=partial(
                api.get_collection_object,
                path, auth=self.auth, conn_url=self.conn_url
            ),
            get_link=partial(
                api.get_collection_object_presigned_url,
                path, auth=self.auth, conn_url=self.conn_url
            ),
            post_link=partial(
                api.post_collection_object_presigned_url,
                path, auth=self.auth, conn_url=self.conn_url
            ),
            list=partial(
                api.get_collection_objects,
                path, auth=self.auth, conn_url=self.conn_url
            ),
        )




import hashlib
import inspect
import os
import pickle
from typing import Iterable, overload
import io
import dill
import tempfile

import malevich.annotations
from malevich._autoflow import tracer as gn  # engine
from malevich.models import AssetNode, AssetOverride, PythonString


class AssetFactory:
    """Creates binary collections (assets) from files or folders"""

    @staticmethod
    def override(
        path: str | None,
        file: os.PathLike | None = None,
        folder: os.PathLike | None = None,
        files: Iterable[os.PathLike] | None = None,
    ) -> AssetOverride:
        _ = (
            file is not None,
            folder is not None,
            files is not None,
        ).count(True)
        if _ != 1:
            raise ValueError(
                'Exactly one of `file`, `folder`, or `files` must be provided'
            )

        return AssetOverride(
            path=path,
            file=file,
            folder=folder,
            files=files,
        )

    @staticmethod
    def capture_function(
        function: callable,
    ) -> malevich.annotations.Asset:
        tfile_ = tempfile.NamedTemporaryFile(delete=False)
        tfile_.write(dill.dumps((function, inspect.getsource(function)), recurse=True))

        return gn.traced(AssetNode(
            name=function.__name__,
            core_path='__functions__/' +
            function.__name__ + '_' + str(hash(function)),
            real_path=tfile_.name,
            alias='captureof_' + function.__name__,
        ))

    @staticmethod
    def from_remote_path(
        reverse_id: PythonString,
        remote_path: str,
        alias: str | None = None,
    ) -> malevich.annotations.Asset:
        return gn.traced(AssetNode(
            name=reverse_id,
            alias=alias,
            core_path=remote_path,
        ))

    @staticmethod
    @overload
    def from_files(
        reverse_id: PythonString,
        *,
        file: os.PathLike = ...,
        remote_path: str | None = ...,
        alias: str | None = ...,
    ) -> malevich.annotations.Asset:
        pass

    @staticmethod
    @overload
    def from_files(
        reverse_id: PythonString,
        *,
        folder: os.PathLike = ...,
        remote_path: str | None = ...,
        alias: str | None = ...,
    ) -> malevich.annotations.Asset:
        pass

    @staticmethod
    @overload
    def from_files(
        reverse_id: PythonString,
        *,
        files: Iterable[os.PathLike] = ...,
        remote_path: str | None = ...,
        alias: str | None = ...,
    ) -> malevich.annotations.Asset:
        pass

    @staticmethod
    def from_files(
        reverse_id: str,
        *,
        file: os.PathLike | None = None,
        folder: os.PathLike | None = None,
        files: Iterable[os.PathLike] | None = None,
        remote_path: str | None = None,
        alias: str | None = None,
    ) -> malevich.annotations.Asset:
        _ = (
            file is not None,
            folder is not None,
            files is not None,
        ).count(True)
        if _ != 1:
            raise ValueError(
                'Exactly one of `file`, `folder`, or `files` must be provided'
            )

        if file is not None:
            if not os.path.isfile(file):
                raise FileNotFoundError(f'File not found: {file}')

            if not remote_path:
                remote_path = os.path.basename(file)
            return gn.traced(AssetNode(
                name=reverse_id,
                real_path=file,
                alias=alias,
                core_path=remote_path,
            ))

        if folder is not None:
            if not os.path.isdir(folder):
                raise FileNotFoundError(f'Folder not found: {folder}')

            files = []
            for root, _, filenames in os.walk(folder):
                for filename in filenames:
                    files.append(os.path.join(root, filename))

            if not remote_path:
                remote_path = os.path.basename(folder)

            return gn.traced(AssetNode(
                name=reverse_id,
                real_path=files,
                alias=alias,
                core_path=remote_path,
            ))

        if files is not None:
            for file in files:
                if not os.path.isfile(file):
                    raise FileNotFoundError(f'File not found: {file}')

            if not remote_path:
                raise ValueError(
                    '`remote_path` must be provided '
                    'when using `files=` argument'
                )

            return gn.traced(AssetNode(
                name=reverse_id,
                real_path=files,
                alias=alias,
                core_path=remote_path,
            ))

        raise ValueError(
            'Exactly one of `file`, `folder`, or `files` must be provided'
        )



from typing import Optional, TypeVar

from malevich._autoflow.tracer import traced
from malevich.models import BaseNode

T = TypeVar('T', bound=traced[BaseNode] | BaseNode)

def run(node: T, alias: Optional[str] = None) -> T:
    """Tags a node within a flow with additional metadata

    Args:
        node (T): Operation to run
        alias (Optional[str], optional): Alias of the operation. Defaults to None.
    """

    if alias is not None:
        if isinstance(node, traced):
            node.owner.alias = alias
        else:
            node.alias = alias
    return node



from sqlalchemy import Column, String, Integer, Sequence
from .base import Base

class CoreCredentials(Base):
    __tablename__ = 'core_creds'
    
    id = Column(Integer, Sequence('core_creds_seq'), primary_key=True)
    host = Column(String(256))
    user = Column(String())
    password = Column(String())

   


import hashlib
from functools import cache
from typing import Optional

from malevich.models.python_string import PythonString

from .base import BaseNode


class AssetNode(BaseNode):
    name: PythonString
    core_path: Optional[str] = None
    real_path: Optional[str | list[str]] = None
    is_composite: Optional[bool] = False
    persistent: Optional[bool] = False

    @cache
    def magic(self) -> str:
        if self.persistent:
            return hashlib.sha256(self.core_path).hexdigest()
        else:
            bytes_ = b""
            if self.is_composite:
                for file in self.real_path:
                    with open(file, "rb") as f:
                        bytes_ += f.read()
                else:
                    with open(self.real_path, "rb") as f:
                        bytes_ += f.read()
            bytes_ += self.core_path.encode()
            return hashlib.sha256(bytes_).hexdigest()

    def get_core_path(self) -> str:
        return '$'+self.core_path

    def __hash__(self) -> int:
        return super().__hash__()



import hashlib
import json
from typing import Any, Generic, Type, TypeVar

from pydantic import BaseModel, Field

from .base import BaseNode

DocumentClass = TypeVar('DocumentClass', bound=BaseModel)

class DocumentNode(BaseNode):
    reverse_id: str
    core_id: str | None = None
    document: Any = None
    document_class: Type | Type[dict] | None = Field(None, exclude=True)

    def magic(self) -> str:
        return hashlib.sha256(self.dump_document_json().encode()).hexdigest()

    def dump_document_json(self) -> str:
        if isinstance(self.document, BaseModel):
            data = json.dumps(self.document.model_dump(), sort_keys=True)
        else:
            data = json.dumps(self.document, sort_keys=True)
        return data

    def get_core_path(self) -> str:
        if self.core_id is None:
            raise RuntimeError(
                "Attempt to get core path of a document node without core_id"
            )

        return f"#{self.core_id}"
